MODULE domvvl
   !!======================================================================
   !!                       ***  MODULE domvvl   ***
   !! Ocean : 
   !!======================================================================
   !! History :  2.0  !  2006-06  (B. Levier, L. Marie)  original code
   !!            3.1  !  2009-02  (G. Madec, M. Leclair, R. Benshila)  pure z* coordinate
   !!            3.3  !  2011-10  (M. Leclair) totally rewrote domvvl:
   !!                                          vvl option includes z_star and z_tilde coordinates
   !!            3.6  !  2014-11  (P. Mathiot) add ice shelf capability
   !!                 !  2018-01  (J. Chanut) improve ztilde robustness
   !!----------------------------------------------------------------------
   !!   'key_vvl'                              variable volume
   !!----------------------------------------------------------------------
   !!----------------------------------------------------------------------
   !!   dom_vvl_init     : define initial vertical scale factors, depths and column thickness
   !!   dom_vvl_sf_nxt   : Compute next vertical scale factors
   !!   dom_vvl_sf_swp   : Swap vertical scale factors and update the vertical grid
   !!   dom_vvl_interpol : Interpolate vertical scale factors from one grid point to another
   !!   dom_vvl_rst      : read/write restart file
   !!   dom_vvl_ctl      : Check the vvl options
   !!   dom_vvl_orca_fix : Recompute some area-weighted interpolations of vertical scale factors 
   !!                    : to account for manual changes to e[1,2][u,v] in some Straits 
   !!----------------------------------------------------------------------
   !! * Modules used
   USE oce             ! ocean dynamics and tracers
   USE dom_oce         ! ocean space and time domain
   USE sbc_oce         ! ocean surface boundary condition
   USE in_out_manager  ! I/O manager
   USE iom             ! I/O manager library
   USE restart         ! ocean restart
   USE lib_mpp         ! distributed memory computing library
   USE lbclnk          ! ocean lateral boundary conditions (or mpp link)
   USE wrk_nemo        ! Memory allocation
   USE timing          ! Timing
   USE bdy_oce         ! ocean open boundary conditions
   USE sbcrnf          ! river runoff 

   IMPLICIT NONE
   PRIVATE

   !! * Routine accessibility
   PUBLIC  dom_vvl_init       ! called by domain.F90
   PUBLIC  dom_vvl_sf_nxt     ! called by step.F90
   PUBLIC  dom_vvl_sf_swp     ! called by step.F90
   PUBLIC  dom_vvl_interpol   ! called by dynnxt.F90
   PRIVATE dom_vvl_orca_fix   ! called by dom_vvl_interpol

   !!* Namelist nam_vvl
   LOGICAL , PUBLIC                                      :: ln_vvl_zstar = .FALSE.              ! zstar  vertical coordinate
   LOGICAL , PUBLIC                                      :: ln_vvl_ztilde = .FALSE.             ! ztilde vertical coordinate
   LOGICAL , PUBLIC                                      :: ln_vvl_layer = .FALSE.              ! level  vertical coordinate
   LOGICAL                                               :: ln_vvl_ztilde_as_zstar = .FALSE.    ! ztilde vertical coordinate
   LOGICAL                                               :: ln_vvl_zstar_at_eqtor = .FALSE.     ! revert to zstar at equator
   LOGICAL                                               :: ln_vvl_zstar_on_shelf =.FALSE.      ! revert to zstar on shelves
   LOGICAL                                               :: ln_vvl_kepe =.FALSE.                ! kinetic/potential energy transfer
   LOGICAL                                               :: ln_vvl_adv_fct =.FALSE.             ! Centred thickness advection
   LOGICAL                                               :: ln_vvl_adv_cn2 =.TRUE.              ! FCT thickness advection
   LOGICAL                                               :: ln_vvl_dbg = .FALSE.                ! debug control prints
   LOGICAL                                               :: ln_vvl_ramp = .FALSE.               ! Ramp on interfaces displacement
   LOGICAL                                               :: ln_vvl_lap                          ! Laplacian thickness diffusion
   LOGICAL                                               :: ln_vvl_blp                          ! Bilaplacian thickness diffusion 
   LOGICAL                                               :: ln_vvl_regrid                       ! ensure layer separation 
   LOGICAL                                               :: ll_shorizd=.FALSE.                  ! Use "shelf horizon depths"         
   !                                                                                            ! conservation: not used yet
   REAL(wp)                                              :: rn_ahe3_lap               ! thickness diffusion coefficient (Laplacian)
   REAL(wp)                                              :: rn_ahe3_blp               ! thickness diffusion coefficient (Bilaplacian)
   REAL(wp)                                              :: rn_rst_e3t                ! ztilde to zstar restoration timescale [days]
   REAL(wp)                                              :: rn_lf_cutoff              ! cutoff frequency for low-pass filter  [days]
   REAL(wp)                                              :: rn_day_ramp               ! Duration of linear ramp  [days]
   REAL(wp)                                              :: rn_zdef_max               ! maximum fractional e3t deformation
   REAL(wp)                                              :: hsmall=0.01               ! small thickness

   !! * Module variables
   REAL(wp), PUBLIC, ALLOCATABLE, SAVE, DIMENSION(:,:,:) :: un_td, vn_td              ! thickness diffusion transport
   REAL(wp)        , ALLOCATABLE, SAVE, DIMENSION(:,:,:) :: un_lf, vn_lf, hdivn_lf    ! 1st order filtered arrays   
   REAL(wp), PUBLIC, ALLOCATABLE, SAVE, DIMENSION(:,:,:) :: tilde_e3t_b, tilde_e3t_n  ! baroclinic scale factors
   REAL(wp), PUBLIC, ALLOCATABLE, SAVE, DIMENSION(:,:,:) :: tilde_e3t_a, dtilde_e3t_a ! baroclinic scale factors
   REAL(wp)        , ALLOCATABLE, SAVE, DIMENSION(:,:)   :: frq_rst_e3t               ! restoring period for scale factors
   REAL(wp)        , ALLOCATABLE, SAVE, DIMENSION(:,:)   :: frq_rst_hdv               ! restoring period for low freq. divergence
   REAL(wp)        , ALLOCATABLE, SAVE, DIMENSION(:,:)   :: tildemask                 ! mask tilde tendency
   REAL(wp)        , ALLOCATABLE, SAVE, DIMENSION(:,:)   :: hsm, dsm                  ! 
   INTEGER         , ALLOCATABLE, SAVE, DIMENSION(:,:)   :: i_int_bot

   !! * Substitutions
#  include "domzgr_substitute.h90"
#  include "vectopt_loop_substitute.h90"
   !!----------------------------------------------------------------------
   !! NEMO/OPA 3.3 , NEMO-Consortium (2010) 
   !! $Id$
   !! Software governed by the CeCILL licence     (NEMOGCM/NEMO_CeCILL.txt)
   !!----------------------------------------------------------------------

CONTAINS

   INTEGER FUNCTION dom_vvl_alloc()
      !!----------------------------------------------------------------------
      !!                ***  FUNCTION dom_vvl_alloc  ***
      !!----------------------------------------------------------------------
      IF( ln_vvl_zstar ) dom_vvl_alloc = 0
      IF( ln_vvl_ztilde .OR. ln_vvl_layer ) THEN
         ALLOCATE( tilde_e3t_b(jpi,jpj,jpk)  , tilde_e3t_n(jpi,jpj,jpk) , tilde_e3t_a(jpi,jpj,jpk) ,   &
            &      dtilde_e3t_a(jpi,jpj,jpk) , un_td  (jpi,jpj,jpk)     , vn_td  (jpi,jpj,jpk)     ,   &
            &      i_int_bot(jpi,jpj), tildemask(jpi,jpj), hsm(jpi,jpj), dsm(jpi,jpj), &
            &      STAT = dom_vvl_alloc        )
         IF( lk_mpp             )   CALL mpp_sum ( dom_vvl_alloc )
         IF( dom_vvl_alloc /= 0 )   CALL ctl_warn('dom_vvl_alloc: failed to allocate arrays')
         un_td = 0.0_wp
         vn_td = 0.0_wp
      ENDIF
      IF( ln_vvl_ztilde ) THEN
         ALLOCATE( frq_rst_e3t(jpi,jpj) , frq_rst_hdv(jpi,jpj),  &
            &        hdivn_lf(jpi,jpj,jpk),  &
            &           un_lf(jpi,jpj,jpk),  &
            &           vn_lf(jpi,jpj,jpk),  STAT= dom_vvl_alloc )
         IF( lk_mpp             )   CALL mpp_sum ( dom_vvl_alloc )
         IF( dom_vvl_alloc /= 0 )   CALL ctl_warn('dom_vvl_alloc: failed to allocate arrays')
      ENDIF

   END FUNCTION dom_vvl_alloc


   SUBROUTINE dom_vvl_init
      !!----------------------------------------------------------------------
      !!                ***  ROUTINE dom_vvl_init  ***
      !!                   
      !! ** Purpose :  Initialization of all scale factors, depths
      !!               and water column heights
      !!
      !! ** Method  :  - use restart file and/or initialize
      !!               - interpolate scale factors
      !!
      !! ** Action  : - fse3t_(n/b) and tilde_e3t_(n/b)
      !!              - Regrid: fse3(u/v)_n
      !!                        fse3(u/v)_b       
      !!                        fse3w_n           
      !!                        fse3(u/v)w_b      
      !!                        fse3(u/v)w_n      
      !!                        fsdept_n, fsdepw_n and fsde3w_n
      !!              - h(t/u/v)_0
      !!              - frq_rst_e3t and frq_rst_hdv
      !!
      !! Reference  : Leclair, M., and G. Madec, 2011, Ocean Modelling.
      !!----------------------------------------------------------------------
      USE phycst,  ONLY : rpi, rsmall, rad
      !! * Local declarations
      INTEGER ::   ji, jj, jk
      INTEGER ::   ii0, ii1, ij0, ij1
      REAL(wp)::   zcoef, zwgt, ztmp, zhmin, zhmax
      !!----------------------------------------------------------------------
      IF( nn_timing == 1 )  CALL timing_start('dom_vvl_init')

      IF(lwp) WRITE(numout,*)
      IF(lwp) WRITE(numout,*) 'dom_vvl_init : Variable volume activated'
      IF(lwp) WRITE(numout,*) '~~~~~~~~~~~~'

      ! choose vertical coordinate (z_star, z_tilde or layer)
      ! ==========================
      CALL dom_vvl_ctl

      ! Allocate module arrays
      ! ======================
      IF( dom_vvl_alloc() /= 0 )   CALL ctl_stop( 'STOP', 'dom_vvl_init : unable to allocate arrays' )

      ! Read or initialize fse3t_(b/n), tilde_e3t_(b/n) and hdiv_lf (and e3t_a(jpk))
      ! ============================================================================
      CALL dom_vvl_rst( nit000, 'READ' )
      fse3t_a(:,:,jpk) = e3t_0(:,:,jpk)

      ! Reconstruction of all vertical scale factors at now and before time steps
      ! =============================================================================
      ! Horizontal scale factor interpolations
      ! --------------------------------------
      CALL dom_vvl_interpol( fse3t_b(:,:,:), fse3u_b(:,:,:), 'U' )
      CALL dom_vvl_interpol( fse3t_b(:,:,:), fse3v_b(:,:,:), 'V' )
      CALL dom_vvl_interpol( fse3t_n(:,:,:), fse3u_n(:,:,:), 'U' )
      CALL dom_vvl_interpol( fse3t_n(:,:,:), fse3v_n(:,:,:), 'V' )
      CALL dom_vvl_interpol( fse3t_n(:,:,:), fse3f_n(:,:,:), 'F' )
      ! Vertical scale factor interpolations
      ! ------------------------------------
      CALL dom_vvl_interpol( fse3t_n(:,:,:), fse3w_n (:,:,:), 'W'  )
      CALL dom_vvl_interpol( fse3u_n(:,:,:), fse3uw_n(:,:,:), 'UW' )
      CALL dom_vvl_interpol( fse3v_n(:,:,:), fse3vw_n(:,:,:), 'VW' )
      CALL dom_vvl_interpol( fse3t_b(:,:,:), fse3w_b (:,:,:), 'W'  )
      CALL dom_vvl_interpol( fse3u_b(:,:,:), fse3uw_b(:,:,:), 'UW' )
      CALL dom_vvl_interpol( fse3v_b(:,:,:), fse3vw_b(:,:,:), 'VW' )
      ! t- and w- points depth
      ! ----------------------
      ! set the isf depth as it is in the initial step
      fsdept_n(:,:,1) = 0.5_wp * fse3w_n(:,:,1)
      fsdepw_n(:,:,1) = 0.0_wp
      fsde3w_n(:,:,1) = fsdept_n(:,:,1) - sshn(:,:)
      fsdept_b(:,:,1) = 0.5_wp * fse3w_b(:,:,1)
      fsdepw_b(:,:,1) = 0.0_wp

      DO jk = 2, jpk
         DO jj = 1,jpj
            DO ji = 1,jpi
              !    zcoef = (tmask(ji,jj,jk) - wmask(ji,jj,jk))   ! 0 everywhere tmask = wmask, ie everywhere expect at jk = mikt
                                                     ! 1 everywhere from mbkt to mikt + 1 or 1 (if no isf)
                                                     ! 0.5 where jk = mikt  
               zcoef = (tmask(ji,jj,jk) - wmask(ji,jj,jk))
               fsdepw_n(ji,jj,jk) = fsdepw_n(ji,jj,jk-1) + fse3t_n(ji,jj,jk-1)
               fsdept_n(ji,jj,jk) =      zcoef  * ( fsdepw_n(ji,jj,jk  ) + 0.5 * fse3w_n(ji,jj,jk))  &
                   &                + (1-zcoef) * ( fsdept_n(ji,jj,jk-1) +       fse3w_n(ji,jj,jk)) 
               fsde3w_n(ji,jj,jk) = fsdept_n(ji,jj,jk) - sshn(ji,jj)
               fsdepw_b(ji,jj,jk) = fsdepw_b(ji,jj,jk-1) + fse3t_b(ji,jj,jk-1)
               fsdept_b(ji,jj,jk) =      zcoef  * ( fsdepw_b(ji,jj,jk  ) + 0.5 * fse3w_b(ji,jj,jk))  &
                   &                + (1-zcoef) * ( fsdept_b(ji,jj,jk-1) +       fse3w_b(ji,jj,jk)) 
            END DO
         END DO
      END DO

      ! Before depth and Inverse of the local depth of the water column at u- and v- points
      ! -----------------------------------------------------------------------------------
      hu_b(:,:) = 0.
      hv_b(:,:) = 0.
      DO jk = 1, jpkm1
         hu_b(:,:) = hu_b(:,:) + fse3u_b(:,:,jk) * umask(:,:,jk)
         hv_b(:,:) = hv_b(:,:) + fse3v_b(:,:,jk) * vmask(:,:,jk)
      END DO
      hur_b(:,:) = umask_i(:,:) / ( hu_b(:,:) + 1._wp - umask_i(:,:) )
      hvr_b(:,:) = vmask_i(:,:) / ( hv_b(:,:) + 1._wp - vmask_i(:,:) )

      ! Restoring frequencies for z_tilde coordinate
      ! ============================================
      tildemask(:,:) = 1._wp

      IF( ln_vvl_ztilde ) THEN
         ! Values in days provided via the namelist; use rsmall to avoid possible division by zero errors with faulty settings
         frq_rst_e3t(:,:) = 2.0_wp * rpi / ( MAX( rn_rst_e3t  , rsmall ) * 86400.0_wp )
         frq_rst_hdv(:,:) = 2.0_wp * rpi / ( MAX( rn_lf_cutoff, rsmall ) * 86400.0_wp )
         !
         IF( ln_vvl_ztilde_as_zstar ) THEN
            ! Ignore namelist settings and use these next two to emulate z-star using z-tilde
            frq_rst_e3t(:,:) = 0.0_wp 
            frq_rst_hdv(:,:) = 1.0_wp / rdt
            tildemask(:,:) = 0._wp
         ENDIF
         
         IF ( ln_vvl_zstar_at_eqtor ) THEN
            DO jj = 1, jpj
               DO ji = 1, jpi
                  IF( ABS(gphit(ji,jj)) >= 6.) THEN
                     ! values outside the equatorial band and transition zone (ztilde)
                     frq_rst_e3t(ji,jj) =  2.0_wp * rpi / ( MAX( rn_rst_e3t  , rsmall ) * 86400.e0_wp )
!                     frq_rst_hdv(ji,jj) =  2.0_wp * rpi / ( MAX( rn_lf_cutoff, rsmall ) * 86400.e0_wp )
              
                  ELSEIF( ABS(gphit(ji,jj)) <= 2.5) THEN
                     ! values inside the equatorial band (ztilde as zstar)
                     frq_rst_e3t(ji,jj) =  0.0_wp
!                     frq_rst_hdv(ji,jj) =  1.0_wp / rdt
                     tildemask(ji,jj) = 0._wp
                  ELSE
                     ! values in the transition band (linearly vary from ztilde to ztilde as zstar values)
                     frq_rst_e3t(ji,jj) = 0.0_wp + (frq_rst_e3t(ji,jj)-0.0_wp)*0.5_wp   &
                        &            * (  1.0_wp - COS( rad*(ABS(gphit(ji,jj))-2.5_wp)  &
                        &                                          * 180._wp / 3.5_wp ) )
!                     frq_rst_hdv(ji,jj) = (1.0_wp / rdt)                                &
!                        &            + (  frq_rst_hdv(ji,jj)-(1.e0_wp / rdt) )*0.5_wp   &
!                        &            * (  1._wp  - COS( rad*(ABS(gphit(ji,jj))-2.5_wp)  &
!                        &                                          * 180._wp / 3.5_wp ) )
                     tildemask(ji,jj) = 0.5_wp * (  1._wp  - COS( rad*(ABS(gphit(ji,jj))-2.5_wp)  &
                        &                                                 * 180._wp / 3.5_wp ) )
                  ENDIF
               END DO
            END DO
            IF( cp_cfg == "orca" .AND. jp_cfg == 3 ) THEN
               ii0 = 103   ;   ii1 = 111        ! Suppress ztilde in the Foxe Basin for ORCA2
               ij0 = 128   ;   ij1 = 135   ;   
               frq_rst_e3t( mi0(ii0):mi1(ii1) , mj0(ij0):mj1(ij1) ) =  0.0_wp
               tildemask( mi0(ii0):mi1(ii1) , mj0(ij0):mj1(ij1) )   = 0.0_wp
!               frq_rst_hdv( mi0(ii0):mi1(ii1) , mj0(ij0):mj1(ij1) ) =  1.e0_wp / rdt
            ENDIF
         ENDIF
         !
         IF ( ln_vvl_zstar_on_shelf ) THEN
            zhmin =  50._wp
            zhmax = 100._wp
            DO jj = 1, jpj
               DO ji = 1, jpi
                  zwgt = 1._wp
                  IF(( ht_0(ji,jj)>zhmin).AND.(ht_0(ji,jj) <=zhmax)) THEN
                     zwgt = (ht_0(ji,jj)-zhmin)/(zhmax-zhmin)
                  ELSEIF ( ht_0(ji,jj)<=zhmin) THEN
                     zwgt = 0._wp
                  ENDIF
                  frq_rst_e3t(ji,jj) = MIN(frq_rst_e3t(ji,jj), frq_rst_e3t(ji,jj)*zwgt)
                  tildemask(ji,jj)   = MIN(tildemask(ji,jj), zwgt)
               END DO
            END DO
         ENDIF
         !
         ztmp = MAXVAL( frq_rst_hdv(:,:) )
         IF( lk_mpp )   CALL mpp_max( ztmp )                 ! max over the global domain
         !
         IF ( (ztmp*rdt) > 1._wp) CALL ctl_stop( 'dom_vvl_init: rn_lf_cuttoff is too small' )
         !
      ENDIF

      IF( nn_timing == 1 )  CALL timing_stop('dom_vvl_init')

   END SUBROUTINE dom_vvl_init


   SUBROUTINE dom_vvl_sf_nxt( kt, kcall ) 
      !!----------------------------------------------------------------------
      !!                ***  ROUTINE dom_vvl_sf_nxt  ***
      !!                   
      !! ** Purpose :  - compute the after scale factors used in tra_zdf, dynnxt,
      !!                 tranxt and dynspg routines
      !!
      !! ** Method  :  - z_star case:  Repartition of ssh INCREMENT proportionnaly to the level thickness.
      !!               - z_tilde_case: after scale factor increment = 
      !!                                    high frequency part of horizontal divergence
      !!                                  + retsoring towards the background grid
      !!                                  + thickness difusion
      !!                               Then repartition of ssh INCREMENT proportionnaly
      !!                               to the "baroclinic" level thickness.
      !!
      !! ** Action  :  - hdiv_lf    : restoring towards full baroclinic divergence in z_tilde case
      !!               - tilde_e3t_a: after increment of vertical scale factor 
      !!                              in z_tilde case
      !!               - fse3(t/u/v)_a
      !!
      !! Reference  : Leclair, M., and Madec, G. 2011, Ocean Modelling.
      !!----------------------------------------------------------------------
      REAL(wp), POINTER, DIMENSION(:,:,:) :: ze3t, ztu, ztv
      REAL(wp), POINTER, DIMENSION(:,:  ) :: zht, z_scale, zwu, zwv, zhdiv
      !! * Arguments
      INTEGER, INTENT( in )                  :: kt                    ! time step
      INTEGER, INTENT( in ), OPTIONAL        :: kcall                 ! optional argument indicating call sequence
      !! * Local declarations
      INTEGER                                :: ji, jj, jk            ! dummy loop indices
      INTEGER                                :: ib, ib_bdy, ip, jp    !   "     "     "
      INTEGER , DIMENSION(3)                 :: ijk_max, ijk_min      ! temporary integers
      INTEGER                                :: ncall
      REAL(wp)                               :: z2dt                  ! temporary scalars
      REAL(wp)                               :: z_tmin, z_tmax        ! temporary scalars
      REAL(wp)                               :: zalpha, zwgt          ! temporary scalars
      REAL(wp)                               :: zdu, zdv, zramp
      LOGICAL                                :: ll_do_bclinic         ! temporary logical
      REAL(wp)                               :: ztra, zbtr, ztout, ztin, zfac, zmsku, zmskv
      !!----------------------------------------------------------------------
      IF( nn_timing == 1 )  CALL timing_start('dom_vvl_sf_nxt')
      CALL wrk_alloc( jpi, jpj, zht, z_scale, zwu, zwv, zhdiv )
      CALL wrk_alloc( jpi, jpj, jpk, ze3t, ztu, ztv )

      IF(kt == nit000)   THEN
         IF(lwp) WRITE(numout,*)
         IF(lwp) WRITE(numout,*) 'dom_vvl_sf_nxt : compute after scale factors'
         IF(lwp) WRITE(numout,*) '~~~~~~~~~~~~~~'
      ENDIF

      ll_do_bclinic = .TRUE.
      ncall=1
      IF( PRESENT(kcall) ) THEN
         ! comment line below if tilda coordinate has to be computed at each call
         IF (kcall == 2 .AND. (ln_vvl_ztilde.OR.ln_vvl_layer) ) ll_do_bclinic = .FALSE.
         IF (kcall == 2) ncall=2    
      ENDIF

      IF( neuler == 0 .AND. kt == nit000 ) THEN
         z2dt =  rdt
      ELSE
         z2dt = 2.0_wp * rdt
      ENDIF

      ! ******************************* !
      ! After scale factors at t-points !
      ! ******************************* !

      !                                                ! --------------------------------------------- !
                                                       ! z_star coordinate and barotropic z-tilde part !
      !                                                ! --------------------------------------------- !

      z_scale(:,:) = ( ssha(:,:) - sshb(:,:) ) * ssmask(:,:) / ( ht_0(:,:) + sshn(:,:) + 1._wp - ssmask(:,:) )
      DO jk = 1, jpkm1
         ! formally this is the same as fse3t_a = e3t_0*(1+ssha/ht_0)
         fse3t_a(:,:,jk) = fse3t_b(:,:,jk) + fse3t_n(:,:,jk) * z_scale(:,:) * tmask(:,:,jk)
      END DO
      
      IF((ln_vvl_ztilde .OR. ln_vvl_layer).AND.ll_do_bclinic ) THEN   ! z_tilde or layer coordinate !
         
         tilde_e3t_a(:,:,:) = 0.0_wp  ! tilde_e3t_a used to store tendency terms
         un_td(:,:,:) = 0.0_wp        ! Transport corrections
         vn_td(:,:,:) = 0.0_wp

         zhdiv(:,:) = 0.
         DO jk = 1, jpkm1
            zhdiv(:,:) = zhdiv(:,:) + fse3t_n(:,:,jk) * hdivn(:,:,jk)
         END DO
         zhdiv(:,:) = zhdiv(:,:) / ( ht_0(:,:) + sshn(:,:) + 1._wp - tmask_i(:,:) ) 

         ze3t(:,:,:) = 0._wp
         IF( ln_rnf ) THEN
            CALL sbc_rnf_div( ze3t )          ! runoffs 
            DO jk=1,jpkm1
               tilde_e3t_a(:,:,jk) = tilde_e3t_a(:,:,jk) - ze3t(:,:,jk) * fse3t_n(:,:,jk)
            END DO   
         ENDIF 

         ! Thickness advection:
         ! --------------------
         ! Set advection velocities and source term
         IF ( ln_vvl_ztilde ) THEN
            IF ( ncall==1 ) THEN
               zalpha  = rdt * 2.0_wp * rpi / ( MAX( rn_lf_cutoff, rsmall ) * 86400.0_wp )
               DO jk = 1, jpkm1
                  ztu(:,:,jk) = un(:,:,jk) * fse3u_n(:,:,jk) * e2u(:,:)
                  ztv(:,:,jk) = vn(:,:,jk) * fse3v_n(:,:,jk) * e1v(:,:)
                  ze3t(:,:,jk) = -tilde_e3t_a(:,:,jk) - fse3t_n(:,:,jk) * zhdiv(:,:)
               END DO
               !
               un_lf(:,:,:)  =    un_lf(:,:,:) * (1._wp - zalpha) + zalpha * ztu(:,:,:)
               vn_lf(:,:,:)  =    vn_lf(:,:,:) * (1._wp - zalpha) + zalpha * ztv(:,:,:)
            hdivn_lf(:,:,:)  = hdivn_lf(:,:,:) * (1._wp - zalpha) + zalpha * ze3t(:,:,:)
!               un_lf(:,:,:)  =    un_lf(:,:,:) * (1._wp - zalpha) + zalpha * un_lf2(:,:,:)
!               vn_lf(:,:,:)  =    vn_lf(:,:,:) * (1._wp - zalpha) + zalpha * vn_lf2(:,:,:)
!            hdivn_lf(:,:,:)  = hdivn_lf(:,:,:) * (1._wp - zalpha) + zalpha * hdivn_lf2(:,:,:)
            ENDIF
            !
            DO jk = 1, jpkm1
               ztu(:,:,jk) = (un(:,:,jk)-un_lf(:,:,jk)/fse3u_n(:,:,jk)*r1_e2u(:,:))*umask(:,:,jk)
               ztv(:,:,jk) = (vn(:,:,jk)-vn_lf(:,:,jk)/fse3v_n(:,:,jk)*r1_e1v(:,:))*vmask(:,:,jk)
               tilde_e3t_a(:,:,jk) =  tilde_e3t_a(:,:,jk) + hdivn_lf(:,:,jk) - frq_rst_e3t(:,:) * tilde_e3t_b(:,:,jk)
            END DO

            !
         ELSEIF ( ln_vvl_layer ) THEN
            !
            DO jk = 1, jpkm1
               ztu(:,:,jk) = un(:,:,jk)
               ztv(:,:,jk) = vn(:,:,jk)
            END DO
            !
         ENDIF
         !
         ! Block fluxes through small layers:
!!         DO jk=1,jpkm1
!!            DO ji = 1, jpi
!!               DO jj= 1, jpj
!!                  zmsku = 0.5_wp * ( 1._wp + SIGN(1._wp, fse3u_n(ji,jj,jk) - hsmall) )
!!                  un_td(ji,jj,jk) = un_td(ji,jj,jk) - (1. - zmsku) * un(ji,jj,jk) * fse3u_n(ji,jj,jk) * e2u(ji,jj)
!!                  ztu(ji,jj,jk) = zmsku * ztu(ji,jj,jk)
!!                  IF ( ln_vvl_ztilde ) un_lf(ji,jj,jk) = zmsku * un_lf(ji,jj,jk)
!!                  !
!!                  zmskv = 0.5_wp * ( 1._wp + SIGN(1._wp, fse3v_n(ji,jj,jk) - hsmall) )
!!                  vn_td(ji,jj,jk) = vn_td(ji,jj,jk) - (1. - zmskv) * vn(ji,jj,jk) * fse3v_n(ji,jj,jk) * e1v(ji,jj)
!!                  ztv(ji,jj,jk) = zmskv * ztv(ji,jj,jk)
!!                  IF ( ln_vvl_ztilde ) vn_lf(ji,jj,jk) = zmskv * vn_lf(ji,jj,jk)
!!               END DO
!!            END DO
!!         END DO      
         !
         ! Do advection
         IF     (ln_vvl_adv_fct) THEN
            CALL dom_vvl_adv_fct( kt, tilde_e3t_a, ztu, ztv )
            !
         ELSEIF (ln_vvl_adv_cn2) THEN
            DO jk = 1, jpkm1
               DO jj = 2, jpjm1
                  DO ji = fs_2, fs_jpim1   ! vector opt.
                     tilde_e3t_a(ji,jj,jk) =   tilde_e3t_a(ji,jj,jk) &
                     & -(  e2u(ji,jj)*fse3u(ji,jj,jk) * ztu(ji,jj,jk) - e2u(ji-1,jj  )*fse3u(ji-1,jj  ,jk) * ztu(ji-1,jj  ,jk)       &
                     &   + e1v(ji,jj)*fse3v(ji,jj,jk) * ztv(ji,jj,jk) - e1v(ji  ,jj-1)*fse3v(ji  ,jj-1,jk) * ztv(ji  ,jj-1,jk)  )    &
                     & / ( e1t(ji,jj) * e2t(ji,jj) )
                  END DO
               END DO
            END DO
         ENDIF
         ! 
         ! Thickness anomaly diffusion:
         ! -----------------------------
         ztu(:,:,:) = 0.0_wp
         ztv(:,:,:) = 0.0_wp

         ze3t(:,:,1) = 0.e0
         DO jj = 1, jpj
            DO ji = 1, jpi
               DO jk = 2, jpk
                  ze3t(ji,jj,jk) =  ze3t(ji,jj,jk-1) + tilde_e3t_b(ji,jj,jk-1) * tmask(ji,jj,jk-1) 
               END DO
            END DO
         END DO

         IF ( ln_vvl_blp ) THEN  ! Bilaplacian
            DO jk = 1, jpkm1
               DO jj = 1, jpjm1                 ! First derivative (gradient)
                  DO ji = 1, fs_jpim1   ! vector opt.                  
!                     ztu(ji,jj,jk) =  umask(ji,jj,jk) * re2u_e1u(ji,jj) &
!                                     &  * ( tilde_e3t_b(ji,jj,jk) - tilde_e3t_b(ji+1,jj  ,jk) )
!                     ztv(ji,jj,jk) =  vmask(ji,jj,jk) * re1v_e2v(ji,jj) & 
!                                     &  * ( tilde_e3t_b(ji,jj,jk) - tilde_e3t_b(ji  ,jj+1,jk) )
                     ztu(ji,jj,jk) =  umask(ji,jj,jk) * re2u_e1u(ji,jj) &
                                     &  * ( ze3t(ji,jj,jk) - ze3t(ji+1,jj  ,jk) )
                     ztv(ji,jj,jk) =  vmask(ji,jj,jk) * re1v_e2v(ji,jj) & 
                                     &  * ( ze3t(ji,jj,jk) - ze3t(ji  ,jj+1,jk) )
                  END DO
               END DO

               DO jj = 2, jpjm1                 ! Second derivative (divergence) time the eddy diffusivity coefficient
                  DO ji = fs_2, fs_jpim1 ! vector opt.
                     zht(ji,jj) =  rn_ahe3_blp * r1_e12t(ji,jj) * (   ztu(ji,jj,jk) - ztu(ji-1,jj,jk)   &
                        &                                           + ztv(ji,jj,jk) - ztv(ji,jj-1,jk)   )
                  END DO
               END DO

#if defined key_bdy
               DO ib_bdy=1, nb_bdy
                  DO ib = 1, idx_bdy(ib_bdy)%nblenrim(1)
                     ji = idx_bdy(ib_bdy)%nbi(ib,1)
                     jj = idx_bdy(ib_bdy)%nbj(ib,1)
                     zht(ji,jj) = 0._wp
                  END DO 
               END DO
#endif
               CALL lbc_lnk( zht, 'T', 1. )     ! Lateral boundary conditions (unchanged sgn)

               DO jj = 1, jpjm1                 ! third derivative (gradient)
                  DO ji = 1, fs_jpim1   ! vector opt.
                     ztu(ji,jj,jk) = umask(ji,jj,jk) * re2u_e1u(ji,jj) * ( zht(ji+1,jj  ) - zht(ji,jj) )
                     ztv(ji,jj,jk) = vmask(ji,jj,jk) * re1v_e2v(ji,jj) * ( zht(ji  ,jj+1) - zht(ji,jj) )
                  END DO
               END DO
            END DO
         ENDIF

         IF ( ln_vvl_lap ) THEN    ! Laplacian
            DO jk = 1, jpkm1                    ! First derivative (gradient)
               DO jj = 1, jpjm1
                  DO ji = 1, fs_jpim1   ! vector opt.                  
!                     zdu =  rn_ahe3_lap * umask(ji,jj,jk) * re2u_e1u(ji,jj) &
!                         &  * ( tilde_e3t_b(ji,jj,jk) - tilde_e3t_b(ji+1,jj  ,jk) )
!                     zdv =  rn_ahe3_lap * vmask(ji,jj,jk) * re1v_e2v(ji,jj) & 
!                         &  * ( tilde_e3t_b(ji,jj,jk) - tilde_e3t_b(ji  ,jj+1,jk) )
                     zdu =  rn_ahe3_lap * umask(ji,jj,jk) * re2u_e1u(ji,jj) &
                         &  * ( ze3t(ji,jj,jk) - ze3t(ji+1,jj  ,jk) )
                     zdv =  rn_ahe3_lap * vmask(ji,jj,jk) * re1v_e2v(ji,jj) & 
                         &  * ( ze3t(ji,jj,jk) - ze3t(ji  ,jj+1,jk) )
                     ztu(ji,jj,jk) = ztu(ji,jj,jk) + zdu
                     ztv(ji,jj,jk) = ztv(ji,jj,jk) + zdv
                  END DO
               END DO
            END DO
         ENDIF 

         ! divergence of diffusive fluxes
         DO jk = 1, jpkm1
            DO jj = 2, jpjm1
               DO ji = fs_2, fs_jpim1   ! vector opt.
!                  tilde_e3t_a(ji,jj,jk) = tilde_e3t_a(ji,jj,jk) + (   ztu(ji-1,jj  ,jk) - ztu(ji,jj,jk)    &
!                     &                                          +     ztv(ji  ,jj-1,jk) - ztv(ji,jj,jk)    &
!                     &                                            ) * r1_e12t(ji,jj)
                  un_td(ji,jj,jk) = un_td(ji,jj,jk) + ztu(ji,jj,jk+1) - ztu(ji,jj,jk  ) 
                  vn_td(ji,jj,jk) = vn_td(ji,jj,jk) + ztv(ji,jj,jk+1) - ztv(ji,jj,jk  ) 
                  tilde_e3t_a(ji,jj,jk) = tilde_e3t_a(ji,jj,jk) + (   ztu(ji-1,jj  ,jk+1) - ztu(ji,jj,jk+1)    &
                     &                                               +ztv(ji  ,jj-1,jk+1) - ztv(ji,jj,jk+1)    &
                     &                                               -ztu(ji-1,jj  ,jk  ) + ztu(ji,jj,jk  )    &
                     &                                               -ztv(ji  ,jj-1,jk  ) + ztv(ji,jj,jk  )    &
                     &                                            ) * r1_e12t(ji,jj)
               END DO
            END DO
         END DO

 
!         un_td(:,:,:) = un_td(:,:,:) + ztu(:,:,:)
!         vn_td(:,:,:) = vn_td(:,:,:) + ztv(:,:,:)

         CALL lbc_lnk( un_td , 'U' , -1.)
         CALL lbc_lnk( vn_td , 'V' , -1.) 
         !
         CALL dom_vvl_ups_cor( kt, tilde_e3t_a, un_td, vn_td )

!         IF ( ln_vvl_ztilde ) THEN
!            ztu(:,:,:) = -un_lf(:,:,:)
!            ztv(:,:,:) = -vn_lf(:,:,:)
!            CALL dom_vvl_ups_cor( kt, tilde_e3t_a, ztu, ztv )
!         ENDIF
         !
         ! Remove "external thickness" tendency:
         DO jk = 1, jpkm1
            tilde_e3t_a(:,:,jk) = tilde_e3t_a(:,:,jk) +   fse3t_n(:,:,jk) * zhdiv(:,:)
         END DO 
                   
         ! Leapfrog time stepping
         ! ~~~~~~~~~~~~~~~~~~~~~~
         zramp = 1._wp
         IF ((.NOT.ln_rstart).AND.ln_vvl_ramp) zramp = MIN(MAX( ((kt-nit000)*rdt)/(rn_day_ramp*rday),0._wp),1._wp)

         DO jk=1,jpkm1
            tilde_e3t_a(:,:,jk) = tilde_e3t_b(:,:,jk) + z2dt * tmask(:,:,jk) * tilde_e3t_a(:,:,jk) &
                               & * tildemask(:,:) * zramp
         END DO

         ! Ensure layer separation:
         ! ~~~~~~~~~~~~~~~~~~~~~~~~
         IF ( ln_vvl_regrid ) CALL dom_vvl_regrid( kt )
  
         ! Boundary conditions:
         ! ~~~~~~~~~~~~~~~~~~~~
#if defined key_bdy
         DO ib_bdy=1, nb_bdy
            DO ib = 1, idx_bdy(ib_bdy)%nblenrim(1)
!!            DO ib = 1, idx_bdy(ib_bdy)%nblen(1)
               ji = idx_bdy(ib_bdy)%nbi(ib,1)
               jj = idx_bdy(ib_bdy)%nbj(ib,1)
               zwgt = idx_bdy(ib_bdy)%nbw(ib,1)
               ip = bdytmask(ji+1,jj  ) - bdytmask(ji-1,jj  )
               jp = bdytmask(ji  ,jj+1) - bdytmask(ji  ,jj-1)
               DO jk = 1, jpkm1  
                  tilde_e3t_a(ji,jj,jk) = 0.e0
!!                  tilde_e3t_a(ji,jj,jk) = tilde_e3t_a(ji,jj,jk) * (1._wp - zwgt)
!!                  tilde_e3t_a(ji,jj,jk) = tilde_e3t_a(ji+ip,jj+jp,jk) * tmask(ji+ip,jj+jp,jk)
               END DO
            END DO 
         END DO
#endif
         CALL lbc_lnk( tilde_e3t_a(:,:,:), 'T', 1. )        

         ! Maximum deformation control
         ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~
         ze3t(:,:,jpk) = 0.0_wp
         DO jk = 1, jpkm1
            ze3t(:,:,jk) = tilde_e3t_a(:,:,jk) / e3t_0(:,:,jk) * tmask(:,:,jk) * tmask_i(:,:)
         END DO
         z_tmax = MAXVAL( ze3t(:,:,:) )
         IF( lk_mpp )   CALL mpp_max( z_tmax )                 ! max over the global domain
         z_tmin = MINVAL( ze3t(:,:,:) )
         IF( lk_mpp )   CALL mpp_min( z_tmin )                 ! min over the global domain
         ! - ML - test: for the moment, stop simulation for too large e3_t variations
         IF( ( z_tmax .GT. rn_zdef_max ) .OR. ( z_tmin .LT. - rn_zdef_max ) ) THEN
            IF( lk_mpp ) THEN
               CALL mpp_maxloc( ze3t, tmask, z_tmax, ijk_max(1), ijk_max(2), ijk_max(3) )
               CALL mpp_minloc( ze3t, tmask, z_tmin, ijk_min(1), ijk_min(2), ijk_min(3) )
            ELSE
               ijk_max = MAXLOC( ze3t(:,:,:) )
               ijk_max(1) = ijk_max(1) + nimpp - 1
               ijk_max(2) = ijk_max(2) + njmpp - 1
               ijk_min = MINLOC( ze3t(:,:,:) )
               ijk_min(1) = ijk_min(1) + nimpp - 1
               ijk_min(2) = ijk_min(2) + njmpp - 1
            ENDIF
            IF (lwp) THEN
               WRITE(numout, *) 'MAX( tilde_e3t_a(:,:,:) / e3t_0(:,:,:) ) =', z_tmax
               WRITE(numout, *) 'at i, j, k=', ijk_max
               WRITE(numout, *) 'MIN( tilde_e3t_a(:,:,:) / e3t_0(:,:,:) ) =', z_tmin
               WRITE(numout, *) 'at i, j, k=', ijk_min            
               CALL ctl_warn('MAX( ABS( tilde_e3t_a(:,:,:) ) / e3t_0(:,:,:) ) too high')
            ENDIF
         ENDIF         
      ENDIF

!      IF( ln_vvl_ztilde )  THEN
!         IF ( ncall==1 ) THEN
!            zalpha  = rdt * 2.0_wp * rpi / ( MAX( rn_lf_cutoff, rsmall ) * 86400.0_wp )
!            DO jk = 1, jpkm1
!               ztu(:,:,jk) = un(:,:,jk) * fse3u_n(:,:,jk) * e2u(:,:) + un_td(:,:,jk)
!               ztv(:,:,jk) = vn(:,:,jk) * fse3v_n(:,:,jk) * e1v(:,:) + vn_td(:,:,jk)
!               ze3t(:,:,jk) = -fse3t_n(:,:,jk) * zhdiv(:,:)
!            END DO
!            !
!               un_lf(:,:,:)  =    un_lf(:,:,:) * (1._wp - zalpha) + zalpha * ztu(:,:,:)
!               vn_lf(:,:,:)  =    vn_lf(:,:,:) * (1._wp - zalpha) + zalpha * ztv(:,:,:)
!            hdivn_lf(:,:,:)  = hdivn_lf(:,:,:) * (1._wp - zalpha) + zalpha * ze3t(:,:,:)
!!               un_lf(:,:,:)  =    un_lf(:,:,:) * (1._wp - zalpha) + zalpha * un_lf2(:,:,:)
!!               vn_lf(:,:,:)  =    vn_lf(:,:,:) * (1._wp - zalpha) + zalpha * vn_lf2(:,:,:)
!!            hdivn_lf(:,:,:)  = hdivn_lf(:,:,:) * (1._wp - zalpha) + zalpha * hdivn_lf2(:,:,:)
!         ENDIF
!      ENDIF

      IF( ln_vvl_ztilde .OR. ln_vvl_layer )  THEN   ! z_tilde or layer coordinate !
      !                                             ! ---baroclinic part--------- !
         DO jk = 1, jpkm1
            fse3t_a(:,:,jk) = fse3t_a(:,:,jk) + (tilde_e3t_a(:,:,jk) - tilde_e3t_b(:,:,jk))                     
         END DO
      ENDIF

      IF(((ln_vvl_ztilde.AND.(.NOT.ln_vvl_ztilde_as_zstar)) .OR. ln_vvl_layer).AND.(.NOT.ll_do_bclinic) ) THEN
         zhdiv(:,:) = 0.
         DO jk = 1, jpkm1
            zhdiv(:,:) = zhdiv(:,:) + fse3t_n(:,:,jk) * (hdivn(:,:,jk) - hdivb(:,:,jk))
         END DO
         zhdiv(:,:) = zhdiv(:,:) / ( ht_0(:,:) + sshn(:,:) + 1. - tmask(:,:,1) )

         IF( neuler == 0 .AND. kt == nit000 ) THEN
            z2dt =  rdt
         ELSE
            z2dt = 2.0_wp * rdt
         ENDIF

         DO jk = 1, jpkm1
            tilde_e3t_a(:,:,jk) = tilde_e3t_a(:,:,jk) - tildemask(:,:) * z2dt * fse3t_n(:,:,jk) * & 
                                  & (hdivn(:,:,jk) - hdivb(:,:,jk) - zhdiv(:,:))
         END DO
         DO jk = 1, jpkm1
            fse3t_a(:,:,jk) = fse3t_a(:,:,jk) - tildemask(:,:) * z2dt * fse3t_n(:,:,jk) * & 
                                  & (hdivn(:,:,jk) - hdivb(:,:,jk) - zhdiv(:,:))
         END DO
      ENDIF


      IF( (ln_vvl_dbg .AND. ( ncall==2 )).AND.(ln_vvl_ztilde .OR. ln_vvl_layer) ) THEN   ! - ML - test: control prints for debuging
         !
         zht(:,:) = 0.0_wp
         DO jk = 1, jpkm1
            zht(:,:) = zht(:,:) + tilde_e3t_a(:,:,jk) * tmask(:,:,jk)
         END DO
         IF( lwp ) WRITE(numout, *) 'kt =', kt
         IF ( ln_vvl_ztilde .OR. ln_vvl_layer ) THEN
            z_tmax = MAXVAL( tmask(:,:,1) * tmask_i(:,:) * ABS( zht(:,:) ), mask = tmask(:,:,1) == 1.e0 )
            IF( lk_mpp ) CALL mpp_max( z_tmax )                             ! max over the global domain
            IF( lwp    ) WRITE(numout, *) kt,' MAXVAL(abs(SUM(tilde_e3t_a))) =', z_tmax
         END IF
         !
         z_tmin = MINVAL( fse3t_n(:,:,:), mask = tmask(:,:,:) == 1.e0  )
         IF( lk_mpp )   CALL mpp_min( z_tmin )                 ! min over the global domain
         IF( lwp    ) WRITE(numout, *) kt,' MINVAL(fse3t_n) =', z_tmin
         IF ( z_tmin .LE. 0._wp ) THEN
            IF( lk_mpp ) THEN
               CALL mpp_minloc(fse3t_n(:,:,:), tmask, z_tmin, ijk_min(1), ijk_min(2), ijk_min(3) )
            ELSE
               ijk_min = MINLOC( fse3t_n(:,:,:), mask = tmask(:,:,:) == 1.e0   )
               ijk_min(1) = ijk_min(1) + nimpp - 1
               ijk_min(2) = ijk_min(2) + njmpp - 1
            ENDIF
            IF (lwp) THEN
               WRITE(numout, *) 'Negative scale factor, fse3t_n =', z_tmin
               WRITE(numout, *) 'at i, j, k=', ijk_min            
               CALL ctl_stop('dom_vvl_sf_nxt: Negative scale factor')
            ENDIF
         ENDIF         
         !
         z_tmin = MINVAL( fse3u_n(:,:,:), mask = umask(:,:,:) == 1.e0 )
         IF( lk_mpp )   CALL mpp_min( z_tmin )                 ! min over the global domain
         IF( lwp    ) WRITE(numout, *) kt,' MINVAL(fse3u_n) =', z_tmin
         IF ( z_tmin .LE. 0._wp ) THEN
            IF( lk_mpp ) THEN
               CALL mpp_minloc(fse3u_n(:,:,:), umask, z_tmin, ijk_min(1), ijk_min(2), ijk_min(3) )
            ELSE
               ijk_min = MINLOC( fse3u_n(:,:,:), mask = umask(:,:,:) == 1.e0   )
               ijk_min(1) = ijk_min(1) + nimpp - 1
               ijk_min(2) = ijk_min(2) + njmpp - 1
            ENDIF
            IF (lwp) THEN
               WRITE(numout, *) 'Negative scale factor, fse3u_n =', z_tmin
               WRITE(numout, *) 'at i, j, k=', ijk_min            
               CALL ctl_stop('dom_vvl_sf_nxt: Negative scale factor')
            ENDIF
         ENDIF 
         !
         z_tmin = MINVAL( fse3v_n(:,:,:), mask = vmask(:,:,:) == 1.e0 )
         IF( lk_mpp )   CALL mpp_min( z_tmin )                 ! min over the global domain
         IF( lwp    ) WRITE(numout, *) kt,' MINVAL(fse3v_n) =', z_tmin
         IF ( z_tmin .LE. 0._wp ) THEN
            IF( lk_mpp ) THEN
               CALL mpp_minloc(fse3v_n(:,:,:), vmask, z_tmin, ijk_min(1), ijk_min(2), ijk_min(3) )
            ELSE
               ijk_min = MINLOC( fse3v_n(:,:,:), mask = vmask(:,:,:) == 1.e0   )
               ijk_min(1) = ijk_min(1) + nimpp - 1
               ijk_min(2) = ijk_min(2) + njmpp - 1
            ENDIF
            IF (lwp) THEN
               WRITE(numout, *) 'Negative scale factor, fse3v_n =', z_tmin
               WRITE(numout, *) 'at i, j, k=', ijk_min            
               CALL ctl_stop('dom_vvl_sf_nxt: Negative scale factor')
            ENDIF
         ENDIF 
         !
         z_tmin = MINVAL( fse3f_n(:,:,:), mask = fmask(:,:,:) == 1.e0 )
         IF( lk_mpp )   CALL mpp_min( z_tmin )                 ! min over the global domain
         IF( lwp    ) WRITE(numout, *) kt,' MINVAL(fse3f_n) =', z_tmin
         IF ( z_tmin .LE. 0._wp ) THEN
            IF( lk_mpp ) THEN
               CALL mpp_minloc(fse3f_n(:,:,:), fmask, z_tmin, ijk_min(1), ijk_min(2), ijk_min(3) )
            ELSE
               ijk_min = MINLOC( fse3f_n(:,:,:), mask = fmask(:,:,:) == 1.e0   )
               ijk_min(1) = ijk_min(1) + nimpp - 1
               ijk_min(2) = ijk_min(2) + njmpp - 1
            ENDIF
            IF (lwp) THEN
               WRITE(numout, *) 'Negative scale factor, fse3f_n =', z_tmin
               WRITE(numout, *) 'at i, j, k=', ijk_min            
               CALL ctl_stop('dom_vvl_sf_nxt: Negative scale factor')
            ENDIF
         ENDIF
         !
         zht(:,:) = 0.0_wp
         DO jk = 1, jpkm1
            zht(:,:) = zht(:,:) + fse3t_n(:,:,jk) * tmask(:,:,jk)
         END DO
         z_tmax = MAXVAL( tmask(:,:,1) * tmask_i(:,:) * ABS( ht_0(:,:) + sshn(:,:) - zht(:,:) ) )
         IF( lk_mpp ) CALL mpp_max( z_tmax )                                ! max over the global domain
         IF( lwp    ) WRITE(numout, *) kt,' MAXVAL(abs(ht_0+sshn  -SUM(fse3t_n))) =', z_tmax
         !
         zht(:,:) = 0.0_wp
         DO jk = 1, jpkm1
            zht(:,:) = zht(:,:) + fse3u_n(:,:,jk) * umask(:,:,jk)
         END DO
         zwu(:,:) = 0._wp
         DO jj = 1, jpjm1
            DO ji = 1, fs_jpim1   ! vector opt.
               zwu(ji,jj) = 0.5_wp * umask(ji,jj,1) * r1_e12u(ji,jj)                             &
                        &          * ( e12t(ji,jj) * sshn(ji,jj) + e12t(ji+1,jj) * sshn(ji+1,jj) )
            END DO
         END DO
         CALL lbc_lnk( zwu(:,:), 'U', 1._wp )
         z_tmax = MAXVAL( umask(:,:,1) * umask_i(:,:) * ABS( hu_0(:,:) + zwu(:,:) - zht(:,:) ) )
         IF( lk_mpp ) CALL mpp_max( z_tmax )                                ! max over the global domain
         IF( lwp    ) WRITE(numout, *) kt,' MAXVAL(abs(hu_0+sshu_n-SUM(fse3u_n))) =', z_tmax
         !
         zht(:,:) = 0.0_wp
         DO jk = 1, jpkm1
            zht(:,:) = zht(:,:) + fse3v_n(:,:,jk) * vmask(:,:,jk)
         END DO
         zwv(:,:) = 0._wp
         DO jj = 1, jpjm1
            DO ji = 1, fs_jpim1   ! vector opt.
               zwv(ji,jj) = 0.5_wp * vmask(ji,jj,1) * r1_e12v(ji,jj)                             &
                        &          * ( e12t(ji,jj) * sshn(ji,jj) + e12t(ji,jj+1) * sshn(ji,jj+1) )
            END DO
         END DO
         CALL lbc_lnk( zwv(:,:), 'V', 1._wp )
         z_tmax = MAXVAL( vmask(:,:,1) * vmask_i(:,:) * ABS( hv_0(:,:) + zwv(:,:) - zht(:,:) ) )
         IF( lk_mpp ) CALL mpp_max( z_tmax )                                ! max over the global domain
         IF( lwp    ) WRITE(numout, *) kt,' MAXVAL(abs(hv_0+sshv_n-SUM(fse3v_n))) =', z_tmax
         !
         zht(:,:) = 0.0_wp
         DO jk = 1, jpkm1
            DO jj = 1, jpjm1
               zht(:,jj) = zht(:,jj) + fse3f_n(:,jj,jk) * umask(:,jj,jk)*umask(:,jj+1,jk)
            END DO
         END DO
         zwu(:,:) = 0._wp
         DO jj = 1, jpjm1
            DO ji = 1, fs_jpim1   ! vector opt.
               zwu(ji,jj) = 0.25_wp * umask(ji,jj,1) * umask(ji,jj+1,1) * r1_e12f(ji,jj)  &
                        &          * (  e12t(ji  ,jj) * sshn(ji  ,jj) + e12t(ji  ,jj+1) * sshn(ji  ,jj+1) &
                        &             + e12t(ji+1,jj) * sshn(ji+1,jj) + e12t(ji+1,jj+1) * sshn(ji+1,jj+1) )
            END DO
         END DO
         CALL lbc_lnk( zht(:,:), 'F', 1._wp )
         CALL lbc_lnk( zwu(:,:), 'F', 1._wp )
         z_tmax = MAXVAL( fmask(:,:,1) * fmask_i(:,:) * ABS( hf_0(:,:) + zwu(:,:) - zht(:,:) ) )
         IF( lk_mpp ) CALL mpp_max( z_tmax )                                ! max over the global domain
         IF( lwp    ) WRITE(numout, *) kt,' MAXVAL(abs(hf_0+sshf_n-SUM(fse3f_n))) =', z_tmax
         !
      END IF

      ! *********************************** !
      ! After scale factors at u- v- points !
      ! *********************************** !

      CALL dom_vvl_interpol( fse3t_a(:,:,:), fse3u_a(:,:,:), 'U' )
      CALL dom_vvl_interpol( fse3t_a(:,:,:), fse3v_a(:,:,:), 'V' )

      ! *********************************** !
      ! After depths at u- v points         !
      ! *********************************** !

      hu_a(:,:) = 0._wp                        ! Ocean depth at U-points
      hv_a(:,:) = 0._wp                        ! Ocean depth at V-points
      DO jk = 1, jpkm1
         hu_a(:,:) = hu_a(:,:) + fse3u_a(:,:,jk) * umask(:,:,jk)
         hv_a(:,:) = hv_a(:,:) + fse3v_a(:,:,jk) * vmask(:,:,jk)
      END DO
      !                                        ! Inverse of the local depth
      hur_a(:,:) = 1._wp / ( hu_a(:,:) + 1._wp - umask_i(:,:) ) * umask_i(:,:)
      hvr_a(:,:) = 1._wp / ( hv_a(:,:) + 1._wp - vmask_i(:,:) ) * vmask_i(:,:)

      CALL wrk_dealloc( jpi, jpj, zht, z_scale, zwu, zwv, zhdiv )
      CALL wrk_dealloc( jpi, jpj, jpk, ze3t, ztu, ztv           )

      IF( nn_timing == 1 )  CALL timing_stop('dom_vvl_sf_nxt')

   END SUBROUTINE dom_vvl_sf_nxt


   SUBROUTINE dom_vvl_sf_swp( kt )
      !!----------------------------------------------------------------------
      !!                ***  ROUTINE dom_vvl_sf_swp  ***
      !!                   
      !! ** Purpose :  compute time filter and swap of scale factors 
      !!               compute all depths and related variables for next time step
      !!               write outputs and restart file
      !!
      !! ** Method  :  - swap of e3t with trick for volume/tracer conservation
      !!               - reconstruct scale factor at other grid points (interpolate)
      !!               - recompute depths and water height fields
      !!
      !! ** Action  :  - fse3t_(b/n), tilde_e3t_(b/n) and fse3(u/v)_n ready for next time step
      !!               - Recompute:
      !!                    fse3(u/v)_b       
      !!                    fse3w_n           
      !!                    fse3(u/v)w_b      
      !!                    fse3(u/v)w_n      
      !!                    fsdept_n, fsdepw_n  and fsde3w_n
      !!                    h(u/v) and h(u/v)r
      !!
      !! Reference  : Leclair, M., and G. Madec, 2009, Ocean Modelling.
      !!              Leclair, M., and G. Madec, 2011, Ocean Modelling.
      !!----------------------------------------------------------------------
      !! * Arguments
      INTEGER, INTENT( in )               :: kt       ! time step
      !! * Local declarations
      REAL(wp), POINTER, DIMENSION(:,:,:) :: zwrk
      INTEGER                             :: jk       ! dummy loop indices
      !!----------------------------------------------------------------------

      IF( nn_timing == 1 )  CALL timing_start('dom_vvl_sf_swp')
      !
      IF( kt == nit000 )   THEN
         IF(lwp) WRITE(numout,*)
         IF(lwp) WRITE(numout,*) 'dom_vvl_sf_swp : - time filter and swap of scale factors'
         IF(lwp) WRITE(numout,*) '~~~~~~~~~~~~~~   - interpolate scale factors and compute depths for next time step'
      ENDIF
      !
      ! Time filter and swap of scale factors
      ! =====================================
      ! - ML - fse3(t/u/v)_b are allready computed in dynnxt.
      IF( ln_vvl_ztilde .OR. ln_vvl_layer ) THEN
         IF( neuler == 0 .AND. kt == nit000 ) THEN
            tilde_e3t_b(:,:,:) = tilde_e3t_n(:,:,:)
         ELSE
            tilde_e3t_b(:,:,:) = tilde_e3t_n(:,:,:) & 
            &         + atfp * ( tilde_e3t_b(:,:,:) - 2.0_wp * tilde_e3t_n(:,:,:) + tilde_e3t_a(:,:,:) )
         ENDIF
         tilde_e3t_n(:,:,:) = tilde_e3t_a(:,:,:)
      ENDIF

      fsdept_b(:,:,:) = fsdept_n(:,:,:)
      fsdepw_b(:,:,:) = fsdepw_n(:,:,:)

      fse3t_n(:,:,:) = fse3t_a(:,:,:)
      fse3u_n(:,:,:) = fse3u_a(:,:,:)
      fse3v_n(:,:,:) = fse3v_a(:,:,:)

      ! Compute all missing vertical scale factor and depths
      ! ====================================================
      ! Horizontal scale factor interpolations
      ! --------------------------------------
      ! - ML - fse3u_b and fse3v_b are allready computed in dynnxt
      ! - JC - hu_b, hv_b, hur_b, hvr_b also
      CALL dom_vvl_interpol( fse3t_n(:,:,:), fse3f_n (:,:,:), 'F'  )
      ! Vertical scale factor interpolations
      ! ------------------------------------
      CALL dom_vvl_interpol( fse3t_n(:,:,:), fse3w_n (:,:,:), 'W'  )
      CALL dom_vvl_interpol( fse3u_n(:,:,:), fse3uw_n(:,:,:), 'UW' )
      CALL dom_vvl_interpol( fse3v_n(:,:,:), fse3vw_n(:,:,:), 'VW' )
      CALL dom_vvl_interpol( fse3t_b(:,:,:), fse3w_b (:,:,:), 'W'  )
      CALL dom_vvl_interpol( fse3u_b(:,:,:), fse3uw_b(:,:,:), 'UW' )
      CALL dom_vvl_interpol( fse3v_b(:,:,:), fse3vw_b(:,:,:), 'VW' )
      ! t- and w- points depth
      ! ----------------------
      fsdept_n(:,:,1) = 0.5_wp * fse3w_n(:,:,1)
      fsdepw_n(:,:,1) = 0.0_wp
      fsde3w_n(:,:,1) = fsdept_n(:,:,1) - sshn(:,:)
      DO jk = 2, jpk
         fsdept_n(:,:,jk) = fsdept_n(:,:,jk-1) + fse3w_n(:,:,jk)
         fsdepw_n(:,:,jk) = fsdepw_n(:,:,jk-1) + fse3t_n(:,:,jk-1)
         fsde3w_n(:,:,jk) = fsdept_n(:,:,jk  ) - sshn   (:,:)
      END DO
      ! Local depth and Inverse of the local depth of the water column at u- and v- points
      ! ----------------------------------------------------------------------------------
      hu (:,:) = hu_a (:,:)
      hv (:,:) = hv_a (:,:)

      ! Inverse of the local depth
      hur(:,:) = hur_a(:,:)
      hvr(:,:) = hvr_a(:,:)

      ! Local depth of the water column at t- points
      ! --------------------------------------------
      ht(:,:) = 0.
      DO jk = 1, jpkm1
         ht(:,:) = ht(:,:) + fse3t_n(:,:,jk) * tmask(:,:,jk)
      END DO

      ! Write additional diagnostics 
      ! ============================
      IF( ln_vvl_ztilde .OR. ln_vvl_layer ) CALL dom_vvl_dia( kt) 

      ! write restart file
      ! ==================
      IF( lrst_oce ) CALL dom_vvl_rst( kt, 'WRITE' )
      !
      IF( nn_timing == 1 )  CALL timing_stop('dom_vvl_sf_swp')

   END SUBROUTINE dom_vvl_sf_swp

   SUBROUTINE dom_vvl_dia( kt )
      !!----------------------------------------------------------------------
      !!                ***  ROUTINE dom_vvl_dia  ***
      !!                   
      !! ** Purpose :  Output some diagnostics in ztilde/zlayer cases
      !!
      !!----------------------------------------------------------------------
      !! * Arguments
      INTEGER, INTENT( in )               :: kt       ! time step
      !! * Local declarations
      INTEGER                             :: ji,jj,jk ! dummy loop indices
      REAL(wp) :: zufim1, zufi, zvfjm1, zvfj, ztmp1, z2dt
      REAL(wp), DIMENSION(4)              :: zr1
      REAL(wp), POINTER, DIMENSION(:,:,:) :: zwdw, zout
      !!----------------------------------------------------------------------
      IF( nn_timing == 1 )  CALL timing_start('dom_vvl_dia')
      !
      CALL wrk_alloc( jpi, jpj, jpk, zwdw, zout )
      !
      ! Compute internal interfaces depths:
      !------------------------------------
      IF ( iom_use("dh_tilde").OR.iom_use("depw_tilde").OR.iom_use("stiff_tilde")) THEN
         zwdw(:,:,1) = 0.e0
         DO jj = 1, jpj
            DO ji = 1, jpi
               DO jk = 2, jpkm1
                  zwdw(ji,jj,jk) =  zwdw(ji,jj,jk-1) + & 
                                 & (tilde_e3t_n(ji,jj,jk-1)+e3t_0(ji,jj,jk-1)) * tmask(ji,jj,jk-1)
               END DO
            END DO
         END DO
      ENDIF
      !
      ! Output interface depth anomaly:
      ! -------------------------------
      IF ( iom_use("depw_tilde") ) CALL iom_put( "depw_tilde", (zwdw(:,:,:)-gdepw_0(:,:,:))*tmask(:,:,:) )
      !
      ! Output grid stiffness (T-points):
      ! ---------------------------------
      IF ( iom_use("stiff_tilde"  ) ) THEN
         zr1(:)   = 0.e0
         zout(:,:,jpk) = 0.e0      
         DO ji = 2, jpim1
            DO jj = 2, jpjm1
               DO jk = 1, jpkm1
                  zr1(1) = umask(ji-1,jj  ,jk) *abs( (zwdw(ji  ,jj  ,jk  )-zwdw(ji-1,jj  ,jk  )  & 
                       &                             +zwdw(ji  ,jj  ,jk+1)-zwdw(ji-1,jj  ,jk+1)) &
                       &                            /(zwdw(ji  ,jj  ,jk  )+zwdw(ji-1,jj  ,jk  )  &
                       &                             -zwdw(ji  ,jj  ,jk+1)-zwdw(ji-1,jj  ,jk+1) + rsmall) )
                  zr1(2) = umask(ji  ,jj  ,jk) *abs( (zwdw(ji+1,jj  ,jk  )-zwdw(ji  ,jj  ,jk  )  &
                       &                             +zwdw(ji+1,jj  ,jk+1)-zwdw(ji  ,jj  ,jk+1)) &
                       &                            /(zwdw(ji+1,jj  ,jk  )+zwdw(ji  ,jj  ,jk  )  &
                       &                             -zwdw(ji+1,jj  ,jk+1)-zwdw(ji  ,jj  ,jk+1) + rsmall) )
                  zr1(3) = vmask(ji  ,jj  ,jk) *abs( (zwdw(ji  ,jj+1,jk  )-zwdw(ji  ,jj  ,jk  )  &
                       &                             +zwdw(ji  ,jj+1,jk+1)-zwdw(ji  ,jj  ,jk+1)) &
                       &                            /(zwdw(ji  ,jj+1,jk  )+zwdw(ji  ,jj  ,jk  )  &
                       &                             -zwdw(ji  ,jj+1,jk+1)-zwdw(ji  ,jj  ,jk+1) + rsmall) )
                  zr1(4) = vmask(ji  ,jj-1,jk) *abs( (zwdw(ji  ,jj  ,jk  )-zwdw(ji  ,jj-1,jk  )  &
                       &                             +zwdw(ji  ,jj  ,jk+1)-zwdw(ji  ,jj-1,jk+1)) &
                       &                            /(zwdw(ji  ,jj  ,jk  )+zwdw(ji  ,jj-1,jk  )  &
                       &                             -zwdw(ji,  jj  ,jk+1)-zwdw(ji  ,jj-1,jk+1) + rsmall) )
                  zout(ji,jj,jk) = MAXVAL(zr1(1:4))
               END DO
            END DO
         END DO

         CALL lbc_lnk( zout, 'T', 1. )
         CALL iom_put( "stiff_tilde", zout(:,:,:) ) 
      END IF
      ! Output Horizontal Laplacian of interfaces depths (W-points):
      ! ------------------------------------------------------------
      IF ( iom_use("dh_tilde")   ) THEN
         !
         zout(:,:,1  )=0._wp
         zout(:,:,jpk)=0._wp
         DO jk = 2, jpkm1
            DO jj = 1, jpjm1
               DO ji = 1, fs_jpim1   ! vector opt.                  
                  ua(ji,jj,jk) =  umask(ji,jj,jk) * re2u_e1u(ji,jj) &
                                  &  * ( zwdw(ji,jj,jk) - zwdw(ji+1,jj  ,jk) )
                  va(ji,jj,jk) =  vmask(ji,jj,jk) * re1v_e2v(ji,jj) & 
                                  &  * ( zwdw(ji,jj,jk) - zwdw(ji  ,jj+1,jk) )
               END DO
            END DO
         END DO
            
         DO jk = 2, jpkm1
            DO jj = 2, jpjm1
               DO ji = fs_2, fs_jpim1   ! vector opt.
                  ztmp1 = ( (ua(ji-1,jj  ,jk) - ua(ji,jj,jk))    &
                     &  +   (va(ji  ,jj-1,jk) - va(ji,jj,jk)) ) * SQRT(r1_e12t(ji,jj))
                  zout(ji,jj,jk) = ABS(ztmp1)*tmask(ji,jj,jk)           
               END DO
            END DO
         END DO 
         ! Mask open boundaries:
#if defined key_bdy
         IF (lk_bdy) THEN
            DO jk = 1, jpkm1
               zout(:,:,jk) = zout(:,:,jk) * bdytmask(:,:)
            END DO
         ENDIF
#endif
         CALL lbc_lnk( zout(:,:,:), 'T', 1. )
         !
         CALL iom_put( "dh_tilde", zout(:,:,:) )
      ENDIF
      !
      ! Output vertical Laplacian of interfaces depths (W-points):
      ! ----------------------------------------------------------
      IF ( iom_use("dz_tilde"  ) ) THEN
         zout(:,:,1  ) = 0._wp
         zout(:,:,jpk) = 0._wp
         DO jk=2,jpkm1
            zout(:,:,jk) = 2._wp*ABS(tilde_e3t_n(:,:,jk)+e3t_0(:,:,jk)-tilde_e3t_n(:,:,jk-1)-e3t_0(:,:,jk-1)) & 
                               &   /(tilde_e3t_n(:,:,jk)+e3t_0(:,:,jk)+tilde_e3t_n(:,:,jk-1)+e3t_0(:,:,jk-1)) &
                               &   * tmask(:,:,jk)
         END DO
         CALL iom_put( "dz_tilde", zout(:,:,:) ) 
      END IF
      !
      !
      ! Output low pass U-velocity:
      ! ---------------------------
      IF ( iom_use("un_lf_tilde"  ).AND.ln_vvl_ztilde ) THEN
         zout(:,:,jpk) = 0.e0  
         DO jk=1,jpkm1
            zout(:,:,jk) = un_lf(:,:,jk)/fse3u_n(:,:,jk)*r1_e2u(:,:)
         END DO
         CALL iom_put( "un_lf_tilde", zout(:,:,:) )
      END IF
      !
      ! Output low pass V-velocity:
      ! ---------------------------
      IF ( iom_use("vn_lf_tilde"  ).AND.ln_vvl_ztilde ) THEN
         zout(:,:,jpk) = 0.e0  
         DO jk=1,jpkm1
            zout(:,:,jk) = vn_lf(:,:,jk)/fse3v_n(:,:,jk)*r1_e1v(:,:)
         END DO
         CALL iom_put( "vn_lf_tilde", zout(:,:,:) )
      END IF
      !
      CALL wrk_dealloc( jpi, jpj, jpk, zwdw, zout )
      !
      IF( nn_timing == 1 )  CALL timing_stop('dom_vvl_dia')
      !
   END SUBROUTINE dom_vvl_dia

   SUBROUTINE dom_vvl_interpol( pe3_in, pe3_out, pout )
      !!---------------------------------------------------------------------
      !!                  ***  ROUTINE dom_vvl__interpol  ***
      !!
      !! ** Purpose :   interpolate scale factors from one grid point to another
      !!
      !! ** Method  :   e3_out = e3_0 + interpolation(e3_in - e3_0)
      !!                - horizontal interpolation: grid cell surface averaging
      !!                - vertical interpolation: simple averaging
      !!----------------------------------------------------------------------
      !! * Arguments
      REAL(wp), DIMENSION(jpi,jpj,jpk), INTENT( in    ) ::  pe3_in     ! input e3 to be interpolated
      REAL(wp), DIMENSION(jpi,jpj,jpk), INTENT( inout ) ::  pe3_out    ! output interpolated e3
      CHARACTER(LEN=*), INTENT( in )                    ::  pout       ! grid point of out scale factors
      !                                                                !   =  'U', 'V', 'W, 'F', 'UW' or 'VW'
      !! * Local declarations
      INTEGER ::   nmet                                                ! horizontal interpolation method
      INTEGER ::   ji, jj, jk                                          ! dummy loop indices
      INTEGER ::   jkbot                                               ! bottom level
      LOGICAL ::   l_is_orca                                           ! local logical
      REAL(wp) :: zmin, zdo, zup, ztap, zsmall
      REAL(wp), POINTER, DIMENSION(:,:)   :: zs                        ! Surface interface depth anomaly
      REAL(wp), POINTER, DIMENSION(:,:,:) :: zw                        ! Interface depth anomaly
      !!----------------------------------------------------------------------
      IF( nn_timing == 1 )  CALL timing_start('dom_vvl_interpol')
         !
      l_is_orca = .FALSE.
      IF( cp_cfg == "orca" .AND. jp_cfg == 2 ) l_is_orca = .TRUE.      ! ORCA R2 configuration - will need to correct some locations

 
!      nmet=0 ! Original method (Surely wrong)
!      nmet= 1 ! Interface interpolation
      nmet=2 ! Internal interfaces interpolation only, spread barotropic increment
! Note that we kept surface weighted interpolation for barotropic increment to be compliant
! with what is done in surface pressure module.

      ztap = 0._wp ! Minimum fraction of T-point thickness at cell interfaces
      zsmall = 1e-3_wp

      IF ( (nmet==1).OR.(nmet==2) ) THEN
         SELECT CASE ( pout )
            !
         CASE( 'U', 'V', 'F' )
            ! Compute interface depth anomaly at T-points
            CALL wrk_alloc( jpi, jpj, jpk, zw )
            CALL wrk_alloc( jpi, jpj, zs )
            !
            zw(:,:,:) =  0._wp    
            !
            DO jk=2,jpk
               zw(:,:,jk) = zw(:,:,jk-1) + pe3_in(:,:,jk-1)*tmask(:,:,jk-1)
            END DO 
            ! Interface depth anomalies:
            DO jk=1,jpkm1
               zw(:,:,jk) = zw(:,:,jk) - zw(:,:,jpk) + ht_0(:,:)
            END DO
            zw(:,:,jpk) = ht_0(:,:)
            !
            IF (nmet==2) THEN        ! Consider "internal" interfaces only
               zs(:,:) = - zw(:,:,1) ! Save surface anomaly (ssh)
               !
               DO jj = 1, jpj
                  DO ji = 1, jpi
                     DO jk=1,jpk
                        zw(ji,jj,jk) = (zw(ji,jj,jk) + zs(ji,jj))                                          &
                                     & * ht_0(ji,jj) / (ht_0(ji,jj) + zs(ji,jj) + 1._wp - tmask(ji,jj,1))  &
                                     & * tmask(ji,jj,jk)
                     END DO
                  END DO
               END DO 
            ENDIF 
            !
         END SELECT
      END IF

      pe3_out(:,:,:) = 0._wp

      SELECT CASE ( pout )
         !               ! ------------------------------------- !
      CASE( 'U' )        ! interpolation from T-point to U-point !
         !               ! ------------------------------------- !     
         ! horizontal surface weighted interpolation
         IF (nmet==0) THEN
            DO jk = 1, jpk
               DO jj = 1, jpjm1
                  DO ji = 1, fs_jpim1   ! vector opt.
                     pe3_out(ji,jj,jk) = 0.5_wp * umask(ji,jj,jk) * r1_e12u(ji,jj)                                   &
                        &                       * (   e12t(ji  ,jj) * ( pe3_in(ji  ,jj,jk) - e3t_0(ji  ,jj,jk) )     &
                        &                           + e12t(ji+1,jj) * ( pe3_in(ji+1,jj,jk) - e3t_0(ji+1,jj,jk) ) )
                  END DO
               END DO
            END DO
         ENDIF
         !
         IF ( (nmet==1).OR.(nmet==2) ) THEN
            DO jj = 1, jpjm1
               DO ji = 1, fs_jpim1   ! vector opt.
                  ! Correction at last level:
                  jkbot = mbku(ji,jj)
                  zdo   = hu_0(ji,jj)
                  DO jk=jkbot,1,-1
                     zup = 0.5_wp * ( zw(ji  ,jj,jk) + zw(ji+1,jj,jk) ) 
                     !
                     ! If there is a step, taper bottom interface:
                     IF ((hu_0(ji,jj) < 0.5_wp * ( ht_0(ji,jj) + ht_0(ji+1,jj) ) ).AND.(zup>zdo)) THEN
                        IF ( ht_0(ji+1,jj) < ht_0(ji,jj) ) THEN
!                           zmin = ztap * pe3_in(ji+1,jj,jk)
                           zmin = ztap * (zw(ji+1,jj,jk+1)-zw(ji+1,jj,jk))
                        ELSE
!                           zmin = ztap * pe3_in(ji  ,jj,jk)
                           zmin = ztap * (zw(ji  ,jj,jk+1)-zw(ji  ,jj,jk))
                        ENDIF
                        zup = MIN(zup, zdo-zmin)
                     ENDIF
                     zup = MIN(zup, zdo-zsmall)
                     pe3_out(ji,jj,jk) = zdo - zup - e3u_0(ji,jj,jk)
                     zdo = zup
                  END DO
               END DO
            END DO
         END IF
         ! 
         IF (nmet==2) THEN           ! Spread sea level anomaly
            DO jj = 1, jpjm1
               DO ji = 1, fs_jpim1   ! vector opt.
                  DO jk=1,jpk
                     pe3_out(ji,jj,jk) =       pe3_out(ji,jj,jk)                                 &
                           &               + ( pe3_out(ji,jj,jk) + e3u_0(ji,jj,jk) )             & 
                           &               / ( hu_0(ji,jj) + 1._wp - umask_i(ji,jj) )            &
                           &               * 0.5_wp * umask(ji,jj,jk) * r1_e12u(ji,jj)           &
                           &               * ( e12t(ji,jj)*zs(ji,jj) + e12t(ji+1,jj)*zs(ji+1,jj) )       
                  END DO
               END DO
            END DO
            !
         ENDIF
         !
         IF( l_is_orca ) CALL dom_vvl_orca_fix( pe3_in, pe3_out, pout )
         ! boundary conditions
         CALL lbc_lnk( pe3_out(:,:,:), 'U', 1._wp )
         pe3_out(:,:,:) = pe3_out(:,:,:) + e3u_0(:,:,:)
         ! 
         IF ( (nmet==1).OR.(nmet==2) ) CALL wrk_dealloc( jpi, jpj, zs )
         IF ( (nmet==1).OR.(nmet==2) ) CALL wrk_dealloc( jpi, jpj, jpk, zw )
         !
         !               ! ------------------------------------- !
      CASE( 'V' )        ! interpolation from T-point to V-point !
         !               ! ------------------------------------- !
         ! horizontal surface weighted interpolation
         IF (nmet==0) THEN
            DO jk = 1, jpk
               DO jj = 1, jpjm1
                  DO ji = 1, fs_jpim1   ! vector opt.
                     pe3_out(ji,jj,jk) = 0.5_wp * vmask(ji,jj,jk) * r1_e12v(ji,jj)                                   &
                        &                       * (   e12t(ji,jj  ) * ( pe3_in(ji,jj  ,jk) - e3t_0(ji,jj  ,jk) )     &
                        &                           + e12t(ji,jj+1) * ( pe3_in(ji,jj+1,jk) - e3t_0(ji,jj+1,jk) ) )
                  END DO
               END DO
            END DO
         ENDIF
         !
         IF ( (nmet==1).OR.(nmet==2) ) THEN
            DO jj = 1, jpjm1
               DO ji = 1, fs_jpim1   ! vector opt.
                  ! Correction at last level:
                  jkbot = mbkv(ji,jj)
                  zdo   = hv_0(ji,jj)
                  DO jk=jkbot,1,-1
                     zup = 0.5_wp * ( zw(ji,jj  ,jk) + zw(ji,jj+1,jk) )
                     !
                     ! If there is a step, taper bottom interface:
                     IF ((hv_0(ji,jj) < 0.5_wp * ( ht_0(ji,jj) + ht_0(ji,jj+1) ) ).AND.(zup>zdo)) THEN
                        IF ( ht_0(ji,jj+1) < ht_0(ji,jj) ) THEN
!                           zmin = ztap * pe3_in(ji,jj+1,jk)
                           zmin = ztap * (zw(ji,jj+1,jk+1)-zw(ji,jj+1,jk))
                        ELSE
!                           zmin = ztap * pe3_in(ji  ,jj,jk)
                           zmin = ztap * (zw(ji,jj  ,jk+1)-zw(ji,jj  ,jk))
                        ENDIF
                        zup = MIN(zup, zdo-zmin)                        
                     ENDIF
                     zup = MIN(zup, zdo-zsmall)
                     pe3_out(ji,jj,jk) = zdo - zup - e3v_0(ji,jj,jk)
                     zdo = zup
                  END DO
               END DO
            END DO
         END IF
         ! 
         IF (nmet==2) THEN           ! Spread sea level anomaly
            !
            DO jj = 1, jpjm1
               DO ji = 1, fs_jpim1   ! vector opt.
                  DO jk=1,jpk
                     pe3_out(ji,jj,jk) =       pe3_out(ji,jj,jk)                                 &
                           &               + ( pe3_out(ji,jj,jk) + e3v_0(ji,jj,jk) )             & 
                           &               / ( hv_0(ji,jj) + 1._wp - vmask_i(ji,jj) )            &
                           &               * 0.5_wp * vmask(ji,jj,jk) * r1_e12v(ji,jj)           &
                           &               * ( e12t(ji,jj)*zs(ji,jj) + e12t(ji,jj+1)*zs(ji,jj+1) )       
                  END DO
               END DO
            END DO
            !
         ENDIF
         !
         IF( l_is_orca ) CALL dom_vvl_orca_fix( pe3_in, pe3_out, pout )
         ! boundary conditions
         CALL lbc_lnk( pe3_out(:,:,:), 'V', 1._wp )
         pe3_out(:,:,:) = pe3_out(:,:,:) + e3v_0(:,:,:)
         ! 
         IF ( (nmet==1).OR.(nmet==2) ) CALL wrk_dealloc( jpi, jpj, zs )
         IF ( (nmet==1).OR.(nmet==2) ) CALL wrk_dealloc( jpi, jpj, jpk, zw )
         !
         !               ! ------------------------------------- !
      CASE( 'F' )        ! interpolation from T-point to F-point !
         !               ! ------------------------------------- !
         ! horizontal surface weighted interpolation
         IF (nmet==0) THEN
            DO jk=1,jpk
               DO jj = 1, jpjm1
                  DO ji = 1, fs_jpim1   ! vector opt.
                     pe3_out(ji,jj,jk) = 0.25_wp * umask(ji,jj,jk) * umask(ji,jj+1,jk) * r1_e12f(ji,jj) &
                           &      * (  e12t(ji  ,jj  ) * ( pe3_in(ji  ,jj  ,jk)-e3t_0(ji  ,jj  ,jk) )   & 
                           &         + e12t(ji  ,jj+1) * ( pe3_in(ji  ,jj+1,jk)-e3t_0(ji  ,jj+1,jk) )   &
                           &         + e12t(ji+1,jj  ) * ( pe3_in(ji+1,jj  ,jk)-e3t_0(ji+1,jj  ,jk) )   & 
                           &         + e12t(ji+1,jj+1) * ( pe3_in(ji+1,jj+1,jk)-e3t_0(ji+1,jj+1,jk) ))                                                      
                  END DO
               END DO
            END DO 
         ENDIF
         !
         IF ( (nmet==1).OR.(nmet==2) ) THEN
            DO jj = 1, jpjm1
               DO ji = 1, fs_jpim1   ! vector opt.
                  ! bottom correction:
                  jkbot = MIN(mbku(ji,jj), mbku(ji,jj+1)) 
                  zdo = hf_0(ji,jj)
                  DO jk=jkbot,1,-1
                     zup =  0.25_wp * (  zw(ji  ,jj  ,jk)  & 
                           &           + zw(ji+1,jj  ,jk)  & 
                           &           + zw(ji  ,jj+1,jk)  & 
                           &           + zw(ji+1,jj+1,jk)  )
                     !
                     ! If there is a step, taper bottom interface:
                     IF ((hf_0(ji,jj) < 0.5_wp * ( hu_0(ji,jj  ) + hu_0(ji,jj+1) ) ).AND.(zup>zdo)) THEN
                        IF ( hu_0(ji,jj+1) < hu_0(ji,jj) ) THEN
                           IF ( ht_0(ji+1,jj+1) < ht_0(ji  ,jj+1) ) THEN
!                              zmin = ztap * pe3_in(ji+1,jj+1,jk)
                              zmin = ztap * (zw(ji+1,jj+1,jk+1)-zw(ji+1,jj+1,jk))
                           ELSE
!                              zmin = ztap * pe3_in(ji  ,jj+1,jk)
                              zmin = ztap * (zw(ji  ,jj+1,jk+1)-zw(ji  ,jj+1,jk))
                           ENDIF
                        ELSE
                           IF ( ht_0(ji+1,jj  ) < ht_0(ji  ,jj  ) ) THEN
!                              zmin = ztap * pe3_in(ji+1,jj  ,jk)
                              zmin = ztap * (zw(ji+1,jj  ,jk+1)-zw(ji+1,jj  ,jk))
                           ELSE
!                              zmin = ztap * pe3_in(ji  ,jj  ,jk)
                              zmin = ztap * (zw(ji  ,jj  ,jk+1)-zw(ji  ,jj  ,jk))
                           ENDIF
                        ENDIF
                        zup = MIN(zup, zdo-zmin)
                     ENDIF
                     zup = MIN(zup, zdo-zsmall)
                     !
                     pe3_out(ji,jj,jk) = zdo - zup - e3f_0(ji,jj,jk)
                     zdo = zup
                  END DO
               END DO
            END DO
         ENDIF
         !
         IF (nmet==2) THEN           ! Spread sea level anomaly
            !
            DO jj = 1, jpjm1
               DO ji = 1, fs_jpim1   ! vector opt.
                  DO jk=1,jpk
                     pe3_out(ji,jj,jk) =       pe3_out(ji,jj,jk)                                         &
                           &               + ( pe3_out(ji,jj,jk) + e3f_0(ji,jj,jk) )                     & 
                           &               / ( hf_0(ji,jj) + 1._wp - umask_i(ji,jj)*umask_i(ji,jj+1) )   &
                           &               * 0.25_wp * umask(ji,jj,jk)*umask(ji,jj+1,jk)*r1_e12f(ji,jj)  &
                           &               * ( e12t(ji  ,jj)*zs(ji  ,jj) + e12t(ji  ,jj+1)*zs(ji  ,jj+1) &
                           &                  +e12t(ji+1,jj)*zs(ji+1,jj) + e12t(ji+1,jj+1)*zs(ji+1,jj+1) )               
                  END DO
               END DO
            END DO
         END IF
         !
         IF( l_is_orca ) CALL dom_vvl_orca_fix( pe3_in, pe3_out, pout )
         ! boundary conditions
         CALL lbc_lnk( pe3_out(:,:,:), 'F', 1._wp )
         pe3_out(:,:,:) = pe3_out(:,:,:) + e3f_0(:,:,:)
         ! 
         IF ( (nmet==1).OR.(nmet==2) ) CALL wrk_dealloc( jpi, jpj, zs )
         IF ( (nmet==1).OR.(nmet==2) ) CALL wrk_dealloc( jpi, jpj, jpk, zw )
         !
         !               ! ------------------------------------- !
      CASE( 'W' )        ! interpolation from T-point to W-point !
         !               ! ------------------------------------- !
         ! vertical simple interpolation
         pe3_out(:,:,1) = e3w_0(:,:,1) + pe3_in(:,:,1) - e3t_0(:,:,1)
         ! - ML - The use of mask in this formaula enables the special treatment of the last w- point without indirect adressing
         DO jk = 2, jpk
            pe3_out(:,:,jk) = e3w_0(:,:,jk) + ( 1.0_wp - 0.5_wp * tmask(:,:,jk) ) * ( pe3_in(:,:,jk-1) - e3t_0(:,:,jk-1) )   &
               &                            +            0.5_wp * tmask(:,:,jk)   * ( pe3_in(:,:,jk  ) - e3t_0(:,:,jk  ) )
         END DO
         !               ! -------------------------------------- !
      CASE( 'UW' )       ! interpolation from U-point to UW-point !
         !               ! -------------------------------------- !
         ! vertical simple interpolation
         pe3_out(:,:,1) = e3uw_0(:,:,1) + pe3_in(:,:,1) - e3u_0(:,:,1)
         ! - ML - The use of mask in this formaula enables the special treatment of the last w- point without indirect adressing
         DO jk = 2, jpk
            pe3_out(:,:,jk) = e3uw_0(:,:,jk) + ( 1.0_wp - 0.5_wp * umask(:,:,jk) ) * ( pe3_in(:,:,jk-1) - e3u_0(:,:,jk-1) )   &
               &                             +            0.5_wp * umask(:,:,jk)   * ( pe3_in(:,:,jk  ) - e3u_0(:,:,jk  ) )
         END DO
         !               ! -------------------------------------- !
      CASE( 'VW' )       ! interpolation from V-point to VW-point !
         !               ! -------------------------------------- !
         ! vertical simple interpolation
         pe3_out(:,:,1) = e3vw_0(:,:,1) + pe3_in(:,:,1) - e3v_0(:,:,1)
         ! - ML - The use of mask in this formaula enables the special treatment of the last w- point without indirect adressing
         DO jk = 2, jpk
            pe3_out(:,:,jk) = e3vw_0(:,:,jk) + ( 1.0_wp - 0.5_wp * vmask(:,:,jk) ) * ( pe3_in(:,:,jk-1) - e3v_0(:,:,jk-1) )   &
               &                             +            0.5_wp * vmask(:,:,jk)   * ( pe3_in(:,:,jk  ) - e3v_0(:,:,jk  ) )
         END DO
      END SELECT
      !

      IF( nn_timing == 1 )  CALL timing_stop('dom_vvl_interpol')

   END SUBROUTINE dom_vvl_interpol



   SUBROUTINE dom_vvl_rst( kt, cdrw )
      !!---------------------------------------------------------------------
      !!                   ***  ROUTINE dom_vvl_rst  ***
      !!                     
      !! ** Purpose :   Read or write VVL file in restart file
      !!
      !! ** Method  :   use of IOM library
      !!                if the restart does not contain vertical scale factors,
      !!                they are set to the _0 values
      !!                if the restart does not contain vertical scale factors increments (z_tilde),
      !!                they are set to 0.
      !!----------------------------------------------------------------------
      !! * Arguments
      INTEGER         , INTENT(in) ::   kt     ! ocean time-step
      CHARACTER(len=*), INTENT(in) ::   cdrw   ! "READ"/"WRITE" flag
      !! * Local declarations
      INTEGER ::   jk
      INTEGER, DIMENSION(7) ::   id            ! local integers
      REAL(wp), POINTER, DIMENSION(:,:) :: zhdiv
      !!----------------------------------------------------------------------
      !
      IF( nn_timing == 1 )  CALL timing_start('dom_vvl_rst')
      IF( TRIM(cdrw) == 'READ' ) THEN        ! Read/initialise 
         !                                   ! ===============
         IF( ln_rstart ) THEN                   !* Read the restart file
            CALL rst_read_open                  !  open the restart file if necessary
            CALL iom_get( numror, jpdom_autoglo, 'sshn'   , sshn    )
            !
            id(1) = iom_varid( numror, 'fse3t_b', ldstop = .FALSE. )
            id(2) = iom_varid( numror, 'fse3t_n', ldstop = .FALSE. )
            id(3) = iom_varid( numror, 'tilde_e3t_b', ldstop = .FALSE. )
            id(4) = iom_varid( numror, 'tilde_e3t_n', ldstop = .FALSE. )
            id(5) = iom_varid( numror, 'un_lf'   , ldstop = .FALSE. )
            id(6) = iom_varid( numror, 'vn_lf'   , ldstop = .FALSE. )
            id(7) = iom_varid( numror, 'hdivn_lf', ldstop = .FALSE. )
            !                             ! --------- !
            !                             ! all cases !
            !                             ! --------- !
            IF( MIN( id(1), id(2) ) > 0 ) THEN       ! all required arrays exist
               CALL iom_get( numror, jpdom_autoglo, 'fse3t_b', fse3t_b(:,:,:) )
               CALL iom_get( numror, jpdom_autoglo, 'fse3t_n', fse3t_n(:,:,:) )
               ! needed to restart if land processor not computed 
               IF(lwp) write(numout,*) 'dom_vvl_rst : fse3t_b and fse3t_n found in restart files'
               WHERE ( tmask(:,:,:) == 0.0_wp ) 
                  fse3t_n(:,:,:) = e3t_0(:,:,:)
                  fse3t_b(:,:,:) = e3t_0(:,:,:)
               END WHERE
               IF( neuler == 0 ) THEN
                  fse3t_b(:,:,:) = fse3t_n(:,:,:)
               ENDIF
            ELSE IF( id(1) > 0 ) THEN
               IF(lwp) write(numout,*) 'dom_vvl_rst WARNING : fse3t_n not found in restart files'
               IF(lwp) write(numout,*) 'fse3t_n set equal to fse3t_b.'
               IF(lwp) write(numout,*) 'neuler is forced to 0'
               CALL iom_get( numror, jpdom_autoglo, 'fse3t_b', fse3t_b(:,:,:) )
               fse3t_b(:,:,:) = fse3t_n(:,:,:)
               neuler = 0
            ELSE IF( id(2) > 0 ) THEN
               IF(lwp) write(numout,*) 'dom_vvl_rst WARNING : fse3t_b not found in restart files'
               IF(lwp) write(numout,*) 'fse3t_b set equal to fse3t_n.'
               IF(lwp) write(numout,*) 'neuler is forced to 0'
               CALL iom_get( numror, jpdom_autoglo, 'fse3t_n', fse3t_n(:,:,:) )
               fse3t_b(:,:,:) = fse3t_n(:,:,:)
               neuler = 0
            ELSE
               IF(lwp) write(numout,*) 'dom_vvl_rst WARNING : fse3t_n not found in restart file'
               IF(lwp) write(numout,*) 'Compute scale factor from sshn'
               IF(lwp) write(numout,*) 'neuler is forced to 0'
               DO jk=1,jpk
                  fse3t_n(:,:,jk) =  e3t_0(:,:,jk) * ( ht_0(:,:) + sshn(:,:) ) &
                      &                            / ( ht_0(:,:) + 1._wp - ssmask(:,:) ) * tmask(:,:,jk) &
                      &            + e3t_0(:,:,jk) * (1._wp -tmask(:,:,jk))
               END DO
               fse3t_b(:,:,:) = fse3t_n(:,:,:)
               neuler = 0
            ENDIF
            !                             ! ----------- !
            IF( ln_vvl_zstar ) THEN       ! z_star case !
               !                          ! ----------- !
               IF( MIN( id(3), id(4) ) > 0 ) THEN
                  CALL ctl_stop( 'dom_vvl_rst: z_star cannot restart from a z_tilde or layer run' )
               ENDIF
               !                          ! ----------------------- !
            ELSE                          ! z_tilde and layer cases !
               !                          ! ----------------------- !
               IF( MIN( id(3), id(4) ) > 0 ) THEN  ! all required arrays exist
                  CALL iom_get( numror, jpdom_autoglo, 'tilde_e3t_b', tilde_e3t_b(:,:,:) )
                  CALL iom_get( numror, jpdom_autoglo, 'tilde_e3t_n', tilde_e3t_n(:,:,:) )
               ELSE                            ! one at least array is missing
                  tilde_e3t_b(:,:,:) = 0.0_wp
                  tilde_e3t_n(:,:,:) = 0.0_wp
                  !
                  neuler = 0
               ENDIF                      ! ------------ !
               IF( ln_vvl_ztilde ) THEN   ! z_tilde case !
                  !                       ! ------------ !
                  IF( MINVAL(id(5:7) ) > 0 ) THEN  ! all required arrays exist
                     CALL iom_get( numror, jpdom_autoglo, 'un_lf'   ,    un_lf(:,:,:) )
                     CALL iom_get( numror, jpdom_autoglo, 'vn_lf'   ,    vn_lf(:,:,:) )
                     CALL iom_get( numror, jpdom_autoglo, 'hdivn_lf', hdivn_lf(:,:,:) )
                  ELSE                            ! one at least array is missing
                     un_lf(:,:,:)    = 0.0_wp
                     vn_lf(:,:,:)    = 0.0_wp
                     hdivn_lf(:,:,:) = 0.0_wp
                     neuler = 0
                  ENDIF
               ENDIF
               !
            ENDIF
            !
         ELSE                                   !* Initialize at "rest"
            fse3t_b(:,:,:) = e3t_0(:,:,:)
            fse3t_n(:,:,:) = e3t_0(:,:,:)
            sshn(:,:) = 0.0_wp
            IF( ln_vvl_ztilde .OR. ln_vvl_layer) THEN
               tilde_e3t_b(:,:,:) = 0.0_wp
               tilde_e3t_n(:,:,:) = 0.0_wp
            END IF
            IF( ln_vvl_ztilde ) THEN
               un_lf(:,:,:)    = 0.0_wp
               vn_lf(:,:,:)    = 0.0_wp
               hdivn_lf(:,:,:) = 0.0_wp
            ENDIF
         ENDIF

      ELSEIF( TRIM(cdrw) == 'WRITE' ) THEN   ! Create restart file
         !                                   ! ===================
         IF(lwp) WRITE(numout,*) '---- dom_vvl_rst ----'
         !                                           ! --------- !
         !                                           ! all cases !
         !                                           ! --------- !
         CALL iom_rstput( kt, nitrst, numrow, 'fse3t_b', fse3t_b(:,:,:) )
         CALL iom_rstput( kt, nitrst, numrow, 'fse3t_n', fse3t_n(:,:,:) )
         !                                           ! ----------------------- !
         IF( ln_vvl_ztilde .OR. ln_vvl_layer ) THEN  ! z_tilde and layer cases !
            !                                        ! ----------------------- !
            CALL iom_rstput( kt, nitrst, numrow, 'tilde_e3t_b', tilde_e3t_b(:,:,:) )
            CALL iom_rstput( kt, nitrst, numrow, 'tilde_e3t_n', tilde_e3t_n(:,:,:) )
         END IF
         !                                           ! -------------!    
         IF( ln_vvl_ztilde ) THEN                    ! z_tilde case !
            !                                        ! ------------ !
            CALL iom_rstput( kt, nitrst, numrow, 'un_lf'   , un_lf(:,:,:)    )
            CALL iom_rstput( kt, nitrst, numrow, 'vn_lf'   , vn_lf(:,:,:)    )
         ENDIF

      ENDIF
      IF( nn_timing == 1 )  CALL timing_stop('dom_vvl_rst')

   END SUBROUTINE dom_vvl_rst

   SUBROUTINE dom_vvl_ctl
      !!---------------------------------------------------------------------
      !!                  ***  ROUTINE dom_vvl_ctl  ***
      !!                
      !! ** Purpose :   Control the consistency between namelist options
      !!                for vertical coordinate
      !!----------------------------------------------------------------------
      INTEGER ::   ioptio
      INTEGER ::   ios

      NAMELIST/nam_vvl/ ln_vvl_zstar               , ln_vvl_ztilde                       , &
                      & ln_vvl_layer               , ln_vvl_ztilde_as_zstar              , &
                      & ln_vvl_zstar_at_eqtor      , ln_vvl_zstar_on_shelf               , &
                      & ln_vvl_adv_cn2             , ln_vvl_adv_fct                      , &
                      & ln_vvl_lap                 , ln_vvl_blp                          , &
                      & rn_ahe3_lap                , rn_ahe3_blp                         , &
                      & rn_rst_e3t                 , rn_lf_cutoff                        , &
                      & ln_vvl_regrid              , rn_zdef_max                         , &
                      & ln_vvl_ramp                , rn_day_ramp                         , &
                      & ln_vvl_dbg   ! not yet implemented: ln_vvl_kepe
      !!---------------------------------------------------------------------- 

      REWIND( numnam_ref )              ! Namelist nam_vvl in reference namelist : 
      READ  ( numnam_ref, nam_vvl, IOSTAT = ios, ERR = 901)
901   IF( ios /= 0 ) CALL ctl_nam ( ios , 'nam_vvl in reference namelist', lwp )

      REWIND( numnam_cfg )              ! Namelist nam_vvl in configuration namelist : Parameters of the run
      READ  ( numnam_cfg, nam_vvl, IOSTAT = ios, ERR = 902 )
902   IF( ios /= 0 ) CALL ctl_nam ( ios , 'nam_vvl in configuration namelist', lwp )
      IF(lwm) WRITE ( numond, nam_vvl )

      IF(lwp) THEN                    ! Namelist print
         WRITE(numout,*)
         WRITE(numout,*) 'dom_vvl_ctl : choice/control of the variable vertical coordinate'
         WRITE(numout,*) '~~~~~~~~~~~'
         WRITE(numout,*) '           Namelist nam_vvl : chose a vertical coordinate'
         WRITE(numout,*) '              zstar                      ln_vvl_zstar   = ', ln_vvl_zstar
         WRITE(numout,*) '              ztilde                     ln_vvl_ztilde  = ', ln_vvl_ztilde
         WRITE(numout,*) '              layer                      ln_vvl_layer   = ', ln_vvl_layer
         WRITE(numout,*) '              ztilde as zstar   ln_vvl_ztilde_as_zstar  = ', ln_vvl_ztilde_as_zstar
         ! WRITE(numout,*) '           Namelist nam_vvl : chose kinetic-to-potential energy conservation'
         ! WRITE(numout,*) '                                         ln_vvl_kepe    = ', ln_vvl_kepe
         WRITE(numout,*) '      ztilde near the equator    ln_vvl_zstar_at_eqtor  = ', ln_vvl_zstar_at_eqtor
         WRITE(numout,*) '      ztilde on shelves          ln_vvl_zstar_on_shelf  = ', ln_vvl_zstar_on_shelf
         WRITE(numout,*) '           Namelist nam_vvl : thickness advection scheme'
         WRITE(numout,*) '              2nd order                  ln_vvl_adv_cn2 = ', ln_vvl_adv_cn2
         WRITE(numout,*) '              2nd order FCT              ln_vvl_adv_fct = ', ln_vvl_adv_fct                    
         WRITE(numout,*) '           Namelist nam_vvl : thickness diffusion scheme'
         WRITE(numout,*) '              Laplacian                  ln_vvl_lap     = ', ln_vvl_lap
         WRITE(numout,*) '              Bilaplacian                ln_vvl_blp     = ', ln_vvl_blp
         WRITE(numout,*) '              Laplacian   coefficient    rn_ahe3_lap    = ', rn_ahe3_lap
         WRITE(numout,*) '              Bilaplacian coefficient    rn_ahe3_blp    = ', rn_ahe3_blp
         WRITE(numout,*) '           Namelist nam_vvl : layers regriding'
         WRITE(numout,*) '              ln_vvl_regrid                             = ', ln_vvl_regrid
         WRITE(numout,*) '           Namelist nam_vvl : linear ramp at startup'
         WRITE(numout,*) '              ln_vvl_ramp                               = ', ln_vvl_ramp
         WRITE(numout,*) '              rn_day_ramp                               = ', rn_day_ramp
         WRITE(numout,*) '           Namelist nam_vvl : maximum e3t deformation fractional change'
         WRITE(numout,*) '                                         rn_zdef_max    = ', rn_zdef_max
         IF( ln_vvl_ztilde_as_zstar ) THEN
            WRITE(numout,*) '           ztilde running in zstar emulation mode; '
            WRITE(numout,*) '           ignoring namelist timescale parameters and using:'
            WRITE(numout,*) '                 hard-wired : z-tilde to zstar restoration timescale (days)'
            WRITE(numout,*) '                                         rn_rst_e3t     =    0.0'
            WRITE(numout,*) '                 hard-wired : z-tilde cutoff frequency of low-pass filter (days)'
            WRITE(numout,*) '                                         rn_lf_cutoff   =    1.0/rdt'
         ELSE
            WRITE(numout,*) '           Namelist nam_vvl : z-tilde to zstar restoration timescale (days)'
            WRITE(numout,*) '                                         rn_rst_e3t     = ', rn_rst_e3t
            WRITE(numout,*) '           Namelist nam_vvl : z-tilde cutoff frequency of low-pass filter (days)'
            WRITE(numout,*) '                                         rn_lf_cutoff   = ', rn_lf_cutoff
         ENDIF
         WRITE(numout,*) '           Namelist nam_vvl : debug prints'
         WRITE(numout,*) '                                         ln_vvl_dbg     = ', ln_vvl_dbg
      ENDIF

      ioptio = 0                      ! Parameter control
      IF( ln_vvl_ztilde_as_zstar ) ln_vvl_ztilde = .true.
      IF( ln_vvl_zstar           )        ioptio = ioptio + 1
      IF( ln_vvl_ztilde          )        ioptio = ioptio + 1
      IF( ln_vvl_layer           )        ioptio = ioptio + 1

      IF( ioptio /= 1 )   CALL ctl_stop( 'Choose ONE vertical coordinate in namelist nam_vvl' )

      IF ( ln_vvl_ztilde.OR.ln_vvl_layer ) THEN 
         ioptio = 0                      ! Choose one advection scheme at most
         IF( ln_vvl_adv_cn2         )        ioptio = ioptio + 1
         IF( ln_vvl_adv_fct         )        ioptio = ioptio + 1
         IF( ioptio /= 1 )   CALL ctl_stop( 'Choose ONE thickness advection scheme in namelist nam_vvl' )
      ENDIF

      IF(lwp) THEN                   ! Print the choice
         WRITE(numout,*)
         IF( ln_vvl_zstar           ) WRITE(numout,*) '              zstar vertical coordinate is used'
         IF( ln_vvl_ztilde          ) WRITE(numout,*) '              ztilde vertical coordinate is used'
         IF( ln_vvl_layer           ) WRITE(numout,*) '              layer vertical coordinate is used'
         IF( ln_vvl_ztilde_as_zstar ) WRITE(numout,*) '              to emulate a zstar coordinate'
         ! - ML - Option not developed yet
         ! IF(       ln_vvl_kepe ) WRITE(numout,*) '              kinetic to potential energy transfer : option used'
         ! IF( .NOT. ln_vvl_kepe ) WRITE(numout,*) '              kinetic to potential energy transfer : option not used'
      ENDIF

      ! Use of "shelf horizon depths" should be allowed with s-z coordinates, but we restrict it to zco and zps 
      ! for the time being
      IF ( ln_sco ) THEN 
        ll_shorizd=.FALSE.
      ELSE
        ll_shorizd=.TRUE.
      ENDIF

#if defined key_agrif
      IF (.NOT.Agrif_Root()) CALL ctl_stop( 'AGRIF not implemented with non-linear free surface (key_vvl)' )
#endif

   END SUBROUTINE dom_vvl_ctl

   SUBROUTINE dom_vvl_orca_fix( pe3_in, pe3_out, pout )
      !!---------------------------------------------------------------------
      !!                   ***  ROUTINE dom_vvl_orca_fix  ***
      !!                     
      !! ** Purpose :   Correct surface weighted, horizontally interpolated, 
      !!                scale factors at locations that have been individually
      !!                modified in domhgr. Such modifications break the
      !!                relationship between e12t and e1u*e2u etc.
      !!                Recompute some scale factors ignoring the modified metric.
      !!----------------------------------------------------------------------
      !! * Arguments
      REAL(wp), DIMENSION(jpi,jpj,jpk), INTENT( in    ) ::  pe3_in     ! input e3 to be interpolated
      REAL(wp), DIMENSION(jpi,jpj,jpk), INTENT( inout ) ::  pe3_out    ! output interpolated e3
      CHARACTER(LEN=*), INTENT( in )                    ::  pout       ! grid point of out scale factors
      !                                                                !   =  'U', 'V', 'W, 'F', 'UW' or 'VW'
      !! * Local declarations
      INTEGER ::   ji, jj, jk                                          ! dummy loop indices
      INTEGER ::   ij0, ij1, ii0, ii1                                  ! dummy loop indices
      INTEGER ::   isrow                                               ! index for ORCA1 starting row
      !! acc
      !! Hmm with the time splitting these "fixes" seem to do more harm than good. Temporarily disabled for
      !! the ORCA2 tests (by changing jp_cfg test from 2 to 3) pending further investigations
      !! 
      !                                                ! =====================
      IF( cp_cfg == "orca" .AND. jp_cfg == 3 ) THEN    ! ORCA R2 configuration
         !                                             ! =====================
      !! acc
         IF( nn_cla == 0 ) THEN
            !
            ii0 = 139   ;   ii1 = 140        ! Gibraltar Strait (e2u was modified)
            ij0 = 102   ;   ij1 = 102
            DO jk = 1, jpkm1
               DO jj = mj0(ij0), mj1(ij1)
                  DO ji = mi0(ii0), mi1(ii1)
                     SELECT CASE ( pout )
                     CASE( 'U' )
                        pe3_out(ji,jj,jk) = 0.5_wp * umask(ji,jj,jk)                                        &
                       &                    * (   e1t(ji  ,jj) * ( pe3_in(ji  ,jj,jk) - e3t_0(ji  ,jj,jk) ) &
                       &                    +     e1t(ji+1,jj) * ( pe3_in(ji+1,jj,jk) - e3t_0(ji+1,jj,jk) ) &
                       &                      ) / e1u(ji,jj)   +   e3u_0(ji,jj,jk)
                     CASE( 'F' )
                        pe3_out(ji,jj,jk) = 0.5_wp * umask(ji,jj,jk) * umask(ji,jj+1,jk)                    &
                       &                    * (   e1u(ji  ,jj) * ( pe3_in(ji  ,jj,jk) - e3u_0(ji  ,jj,jk) ) &
                       &                    +     e1u(ji+1,jj) * ( pe3_in(ji+1,jj,jk) - e3u_0(ji+1,jj,jk) ) &
                       &                      ) / e1f(ji,jj)   +   e3f_0(ji,jj,jk)
                     END SELECT
                  END DO
               END DO
            END DO
            !
            ii0 = 160   ;   ii1 = 160        ! Bab el Mandeb (e2u and e1v were modified)
            ij0 =  88   ;   ij1 =  88
            DO jk = 1, jpkm1
               DO jj = mj0(ij0), mj1(ij1)
                  DO ji = mi0(ii0), mi1(ii1)
                     SELECT CASE ( pout )
                     CASE( 'U' )
                        pe3_out(ji,jj,jk) = 0.5_wp * umask(ji,jj,jk)                                        &
                       &                    * (   e1t(ji  ,jj) * ( pe3_in(ji  ,jj,jk) - e3t_0(ji  ,jj,jk) ) &
                       &                    +     e1t(ji+1,jj) * ( pe3_in(ji+1,jj,jk) - e3t_0(ji+1,jj,jk) ) &
                       &                      ) / e1u(ji,jj)   +   e3u_0(ji,jj,jk)
                     CASE( 'V' )
                        pe3_out(ji,jj,jk) = 0.5_wp * vmask(ji,jj,jk)                                        &
                       &                    * (   e2t(ji,jj  ) * ( pe3_in(ji,jj  ,jk) - e3t_0(ji,jj  ,jk) ) &
                       &                    +     e2t(ji,jj+1) * ( pe3_in(ji,jj+1,jk) - e3t_0(ji,jj+1,jk) ) &
                       &                      ) / e2v(ji,jj)   +   e3v_0(ji,jj,jk)
                     CASE( 'F' )
                        pe3_out(ji,jj,jk) = 0.5_wp * umask(ji,jj,jk) * umask(ji,jj+1,jk)                    &
                       &                    * (   e1u(ji  ,jj) * ( pe3_in(ji  ,jj,jk) - e3u_0(ji  ,jj,jk) ) &
                       &                    +     e1u(ji+1,jj) * ( pe3_in(ji+1,jj,jk) - e3u_0(ji+1,jj,jk) ) &
                       &                      ) / e1f(ji,jj)   +   e3f_0(ji,jj,jk)
                     END SELECT
                  END DO
               END DO
            END DO
         ENDIF

         ii0 = 145   ;   ii1 = 146        ! Danish Straits (e2u was modified)
         ij0 = 116   ;   ij1 = 116
         DO jk = 1, jpkm1
            DO jj = mj0(ij0), mj1(ij1)
               DO ji = mi0(ii0), mi1(ii1)
                  SELECT CASE ( pout )
                  CASE( 'U' )
                     pe3_out(ji,jj,jk) = 0.5_wp * umask(ji,jj,jk)                                        &
                    &                    * (   e1t(ji  ,jj) * ( pe3_in(ji  ,jj,jk) - e3t_0(ji  ,jj,jk) ) &
                    &                    +     e1t(ji+1,jj) * ( pe3_in(ji+1,jj,jk) - e3t_0(ji+1,jj,jk) ) &
                    &                      ) / e1u(ji,jj)   +   e3u_0(ji,jj,jk)
                  CASE( 'F' )
                     pe3_out(ji,jj,jk) = 0.5_wp * umask(ji,jj,jk) * umask(ji,jj+1,jk)                    &
                    &                    * (   e1u(ji  ,jj) * ( pe3_in(ji  ,jj,jk) - e3u_0(ji  ,jj,jk) ) &
                    &                    +     e1u(ji+1,jj) * ( pe3_in(ji+1,jj,jk) - e3u_0(ji+1,jj,jk) ) &
                    &                      ) / e1f(ji,jj)   +   e3f_0(ji,jj,jk)
                  END SELECT
               END DO
            END DO
         END DO
      ENDIF
      !
         !                                             ! =====================
      IF( cp_cfg == "orca" .AND. jp_cfg == 1 ) THEN    ! ORCA R1 configuration
         !                                             ! =====================
         ! This dirty section will be suppressed by simplification process:
         ! all this will come back in input files
         ! Currently these hard-wired indices relate to configuration with
         ! extend grid (jpjglo=332)
         ! which had a grid-size of 362x292.
         isrow = 332 - jpjglo
         !
         ii0 = 282           ;   ii1 = 283        ! Gibraltar Strait (e2u was modified)
         ij0 = 241 - isrow   ;   ij1 = 241 - isrow
         DO jk = 1, jpkm1
            DO jj = mj0(ij0), mj1(ij1)
               DO ji = mi0(ii0), mi1(ii1)
                  SELECT CASE ( pout )
                  CASE( 'U' )
                     pe3_out(ji,jj,jk) = 0.5_wp * umask(ji,jj,jk)                                        &
                    &                    * (   e1t(ji  ,jj) * ( pe3_in(ji  ,jj,jk) - e3t_0(ji  ,jj,jk) ) &
                    &                    +     e1t(ji+1,jj) * ( pe3_in(ji+1,jj,jk) - e3t_0(ji+1,jj,jk) ) &
                    &                      ) / e1u(ji,jj)   +   e3u_0(ji,jj,jk)
                  CASE( 'F' )
                     pe3_out(ji,jj,jk) = 0.5_wp * umask(ji,jj,jk) * umask(ji,jj+1,jk)                    &
                    &                    * (   e1u(ji  ,jj) * ( pe3_in(ji  ,jj,jk) - e3u_0(ji  ,jj,jk) ) &
                    &                    +     e1u(ji+1,jj) * ( pe3_in(ji+1,jj,jk) - e3u_0(ji+1,jj,jk) ) &
                    &                      ) / e1f(ji,jj)   +   e3f_0(ji,jj,jk)
                  END SELECT
               END DO
            END DO
         END DO
         !
         ii0 = 314           ;   ii1 = 315        ! Bhosporus Strait (e2u was modified)
         ij0 = 248 - isrow   ;   ij1 = 248 - isrow
         DO jk = 1, jpkm1
            DO jj = mj0(ij0), mj1(ij1)
               DO ji = mi0(ii0), mi1(ii1)
                  SELECT CASE ( pout )
                  CASE( 'U' )
                     pe3_out(ji,jj,jk) = 0.5_wp * umask(ji,jj,jk)                                        &  
                    &                    * (   e1t(ji  ,jj) * ( pe3_in(ji  ,jj,jk) - e3t_0(ji  ,jj,jk) ) &
                    &                    +     e1t(ji+1,jj) * ( pe3_in(ji+1,jj,jk) - e3t_0(ji+1,jj,jk) ) &
                    &                      ) / e1u(ji,jj)   +   e3u_0(ji,jj,jk)
                  CASE( 'F' )
                     pe3_out(ji,jj,jk) = 0.5_wp * umask(ji,jj,jk) * umask(ji,jj+1,jk)                    &  
                    &                    * (   e1u(ji  ,jj) * ( pe3_in(ji  ,jj,jk) - e3u_0(ji  ,jj,jk) ) &
                    &                    +     e1u(ji+1,jj) * ( pe3_in(ji+1,jj,jk) - e3u_0(ji+1,jj,jk) ) &
                    &                      ) / e1f(ji,jj)   +   e3f_0(ji,jj,jk)
                  END SELECT
               END DO
            END DO
         END DO
         !
         ii0 =  44           ;   ii1 =  44        ! Lombok Strait (e1v was modified)
         ij0 = 164 - isrow   ;   ij1 = 165 - isrow
         DO jk = 1, jpkm1
            DO jj = mj0(ij0), mj1(ij1)
               DO ji = mi0(ii0), mi1(ii1)
                  SELECT CASE ( pout )
                  CASE( 'V' )
                     pe3_out(ji,jj,jk) = 0.5_wp * vmask(ji,jj,jk)                                        &
                    &                    * (   e2t(ji,jj  ) * ( pe3_in(ji,jj  ,jk) - e3t_0(ji,jj  ,jk) ) &
                    &                    +     e2t(ji,jj+1) * ( pe3_in(ji,jj+1,jk) - e3t_0(ji,jj+1,jk) ) &
                    &                      ) / e2v(ji,jj)   +   e3v_0(ji,jj,jk)
                  END SELECT
               END DO
            END DO
         END DO
         !
         ii0 =  48           ;   ii1 =  48        ! Sumba Strait (e1v was modified) [closed from bathy_11 on]
         ij0 = 164 - isrow   ;   ij1 = 165 - isrow
         DO jk = 1, jpkm1
            DO jj = mj0(ij0), mj1(ij1)
               DO ji = mi0(ii0), mi1(ii1)
                  SELECT CASE ( pout )
                  CASE( 'V' )
                     pe3_out(ji,jj,jk) = 0.5_wp * vmask(ji,jj,jk)                                        &
                    &                    * (   e2t(ji,jj  ) * ( pe3_in(ji,jj  ,jk) - e3t_0(ji,jj  ,jk) ) &
                    &                    +     e2t(ji,jj+1) * ( pe3_in(ji,jj+1,jk) - e3t_0(ji,jj+1,jk) ) &
                    &                      ) / e2v(ji,jj)   +   e3v_0(ji,jj,jk)
                  END SELECT
               END DO
            END DO
         END DO
         !
         ii0 =  53          ;   ii1 =  53        ! Ombai Strait (e1v was modified)
         ij0 = 164 - isrow  ;   ij1 = 165  - isrow  
         DO jk = 1, jpkm1
            DO jj = mj0(ij0), mj1(ij1)
               DO ji = mi0(ii0), mi1(ii1)
                  SELECT CASE ( pout )
                  CASE( 'V' )
                     pe3_out(ji,jj,jk) = 0.5_wp * vmask(ji,jj,jk)                                        &
                    &                    * (   e2t(ji,jj  ) * ( pe3_in(ji,jj  ,jk) - e3t_0(ji,jj  ,jk) ) &
                    &                    +     e2t(ji,jj+1) * ( pe3_in(ji,jj+1,jk) - e3t_0(ji,jj+1,jk) ) &
                    &                      ) / e2v(ji,jj)   +   e3v_0(ji,jj,jk)
                  END SELECT
               END DO
            END DO
         END DO
         !
         ii0 =  56            ;   ii1 =  56        ! Timor Passage (e1v was modified)
         ij0 = 164 - isrow    ;   ij1 = 165  - isrow  
         DO jk = 1, jpkm1
            DO jj = mj0(ij0), mj1(ij1)
               DO ji = mi0(ii0), mi1(ii1)
                  SELECT CASE ( pout )
                  CASE( 'V' )
                     pe3_out(ji,jj,jk) = 0.5_wp * vmask(ji,jj,jk)                                        &
                    &                    * (   e2t(ji,jj  ) * ( pe3_in(ji,jj  ,jk) - e3t_0(ji,jj  ,jk) ) &
                    &                    +     e2t(ji,jj+1) * ( pe3_in(ji,jj+1,jk) - e3t_0(ji,jj+1,jk) ) &
                    &                      ) / e2v(ji,jj)   +   e3v_0(ji,jj,jk)
                  END SELECT
               END DO
            END DO
         END DO
         !
         ii0 =  55            ;   ii1 =  55        ! West Halmahera Strait (e1v was modified)
         ij0 = 181 - isrow    ;   ij1 = 182 - isrow  
         DO jk = 1, jpkm1
            DO jj = mj0(ij0), mj1(ij1)
               DO ji = mi0(ii0), mi1(ii1)
                  SELECT CASE ( pout )
                  CASE( 'V' )
                     pe3_out(ji,jj,jk) = 0.5_wp * vmask(ji,jj,jk)                                        &
                    &                    * (   e2t(ji,jj  ) * ( pe3_in(ji,jj  ,jk) - e3t_0(ji,jj  ,jk) ) &
                    &                    +     e2t(ji,jj+1) * ( pe3_in(ji,jj+1,jk) - e3t_0(ji,jj+1,jk) ) &
                    &                      ) / e2v(ji,jj)   +   e3v_0(ji,jj,jk)
                  END SELECT
               END DO
            END DO
         END DO
         !
         ii0 =  58            ;   ii1 =  58        ! East Halmahera Strait (e1v was modified)
         ij0 = 181 - isrow    ;   ij1 = 182 - isrow  
         DO jk = 1, jpkm1
            DO jj = mj0(ij0), mj1(ij1)
               DO ji = mi0(ii0), mi1(ii1)
                  SELECT CASE ( pout )
                  CASE( 'V' )
                     pe3_out(ji,jj,jk) = 0.5_wp * vmask(ji,jj,jk)                                        &
                    &                    * (   e2t(ji,jj  ) * ( pe3_in(ji,jj  ,jk) - e3t_0(ji,jj  ,jk) ) &
                    &                    +     e2t(ji,jj+1) * ( pe3_in(ji,jj+1,jk) - e3t_0(ji,jj+1,jk) ) &
                    &                      ) / e2v(ji,jj)   +   e3v_0(ji,jj,jk)
                  END SELECT
               END DO
            END DO
         END DO
      ENDIF
         !                                             ! =====================
      IF( cp_cfg == "orca" .AND. jp_cfg == 05 ) THEN   ! ORCA R05 configuration
         !                                             ! =====================
         !
         ii0 = 563   ;   ii1 = 564        ! Gibraltar Strait (e2u was modified)
         ij0 = 327   ;   ij1 = 327
         DO jk = 1, jpkm1
            DO jj = mj0(ij0), mj1(ij1)
               DO ji = mi0(ii0), mi1(ii1)
                  SELECT CASE ( pout )
                  CASE( 'U' )
                     pe3_out(ji,jj,jk) = 0.5_wp * umask(ji,jj,jk)                                        &
                    &                    * (   e1t(ji  ,jj) * ( pe3_in(ji  ,jj,jk) - e3t_0(ji  ,jj,jk) ) &
                    &                    +     e1t(ji+1,jj) * ( pe3_in(ji+1,jj,jk) - e3t_0(ji+1,jj,jk) ) &
                    &                      ) / e1u(ji,jj)   +   e3u_0(ji,jj,jk)
                  CASE( 'F' )
                     pe3_out(ji,jj,jk) = 0.5_wp * umask(ji,jj,jk) * umask(ji,jj+1,jk)                    &
                    &                    * (   e1u(ji  ,jj) * ( pe3_in(ji  ,jj,jk) - e3u_0(ji  ,jj,jk) ) &
                    &                    +     e1u(ji+1,jj) * ( pe3_in(ji+1,jj,jk) - e3u_0(ji+1,jj,jk) ) &
                    &                      ) / e1f(ji,jj)   +   e3f_0(ji,jj,jk)
                  END SELECT
               END DO
            END DO
         END DO
         !
         ii0 = 627   ;   ii1 = 628        ! Bosphorus Strait (e2u was modified)
         ij0 = 343   ;   ij1 = 343
         DO jk = 1, jpkm1
            DO jj = mj0(ij0), mj1(ij1)
               DO ji = mi0(ii0), mi1(ii1)
                  SELECT CASE ( pout )
                  CASE( 'U' )
                     pe3_out(ji,jj,jk) = 0.5_wp * umask(ji,jj,jk)                                        &  
                    &                    * (   e1t(ji  ,jj) * ( pe3_in(ji  ,jj,jk) - e3t_0(ji  ,jj,jk) ) &
                    &                    +     e1t(ji+1,jj) * ( pe3_in(ji+1,jj,jk) - e3t_0(ji+1,jj,jk) ) &
                    &                      ) / e1u(ji,jj)   +   e3u_0(ji,jj,jk)
                  CASE( 'F' )
                     pe3_out(ji,jj,jk) = 0.5_wp * umask(ji,jj,jk) * umask(ji,jj+1,jk)                    &  
                    &                    * (   e1u(ji  ,jj) * ( pe3_in(ji  ,jj,jk) - e3u_0(ji  ,jj,jk) ) &
                    &                    +     e1u(ji+1,jj) * ( pe3_in(ji+1,jj,jk) - e3u_0(ji+1,jj,jk) ) &
                    &                      ) / e1f(ji,jj)   +   e3f_0(ji,jj,jk)
                  END SELECT
               END DO
            END DO
         END DO
         !
         ii0 =  93   ;   ii1 =  94        ! Sumba Strait (e2u was modified)
         ij0 = 232   ;   ij1 = 232
         DO jk = 1, jpkm1
            DO jj = mj0(ij0), mj1(ij1)
               DO ji = mi0(ii0), mi1(ii1)
                  SELECT CASE ( pout )
                  CASE( 'U' )
                     pe3_out(ji,jj,jk) = 0.5_wp * umask(ji,jj,jk)                                        &
                    &                    * (   e1t(ji  ,jj) * ( pe3_in(ji  ,jj,jk) - e3t_0(ji  ,jj,jk) ) &
                    &                    +     e1t(ji+1,jj) * ( pe3_in(ji+1,jj,jk) - e3t_0(ji+1,jj,jk) ) &
                    &                      ) / e1u(ji,jj)   +   e3u_0(ji,jj,jk)
                  CASE( 'F' )
                     pe3_out(ji,jj,jk) = 0.5_wp * umask(ji,jj,jk) * umask(ji,jj+1,jk)                    &
                    &                    * (   e1u(ji  ,jj) * ( pe3_in(ji  ,jj,jk) - e3u_0(ji  ,jj,jk) ) &
                    &                    +     e1u(ji+1,jj) * ( pe3_in(ji+1,jj,jk) - e3u_0(ji+1,jj,jk) ) &
                    &                      ) / e1f(ji,jj)   +   e3f_0(ji,jj,jk)
                  END SELECT
               END DO
            END DO
         END DO
         !
         ii0 = 103   ;   ii1 = 103        ! Ombai Strait (e2u was modified)
         ij0 = 232   ;   ij1 = 232
         DO jk = 1, jpkm1
            DO jj = mj0(ij0), mj1(ij1)
               DO ji = mi0(ii0), mi1(ii1)
                  SELECT CASE ( pout )
                  CASE( 'U' )
                     pe3_out(ji,jj,jk) = 0.5_wp * umask(ji,jj,jk)                                        &
                    &                    * (   e1t(ji  ,jj) * ( pe3_in(ji  ,jj,jk) - e3t_0(ji  ,jj,jk) ) &
                    &                    +     e1t(ji+1,jj) * ( pe3_in(ji+1,jj,jk) - e3t_0(ji+1,jj,jk) ) &
                    &                      ) / e1u(ji,jj)   +   e3u_0(ji,jj,jk)
                  CASE( 'F' )
                     pe3_out(ji,jj,jk) = 0.5_wp * umask(ji,jj,jk) * umask(ji,jj+1,jk)                    &
                    &                    * (   e1u(ji  ,jj) * ( pe3_in(ji  ,jj,jk) - e3u_0(ji  ,jj,jk) ) &
                    &                    +     e1u(ji+1,jj) * ( pe3_in(ji+1,jj,jk) - e3u_0(ji+1,jj,jk) ) &
                    &                      ) / e1f(ji,jj)   +   e3f_0(ji,jj,jk)
                  END SELECT
               END DO
            END DO
         END DO
         !
         ii0 =  15   ;   ii1 =  15        ! Palk Strait (e2u was modified)
         ij0 = 270   ;   ij1 = 270
         DO jk = 1, jpkm1
            DO jj = mj0(ij0), mj1(ij1)
               DO ji = mi0(ii0), mi1(ii1)
                  SELECT CASE ( pout )
                  CASE( 'U' )
                     pe3_out(ji,jj,jk) = 0.5_wp * umask(ji,jj,jk)                                        &
                    &                    * (   e1t(ji  ,jj) * ( pe3_in(ji  ,jj,jk) - e3t_0(ji  ,jj,jk) ) &
                    &                    +     e1t(ji+1,jj) * ( pe3_in(ji+1,jj,jk) - e3t_0(ji+1,jj,jk) ) &
                    &                      ) / e1u(ji,jj)   +   e3u_0(ji,jj,jk)
                  CASE( 'F' )
                     pe3_out(ji,jj,jk) = 0.5_wp * umask(ji,jj,jk) * umask(ji,jj+1,jk)                    &
                    &                    * (   e1u(ji  ,jj) * ( pe3_in(ji  ,jj,jk) - e3u_0(ji  ,jj,jk) ) &
                    &                    +     e1u(ji+1,jj) * ( pe3_in(ji+1,jj,jk) - e3u_0(ji+1,jj,jk) ) &
                    &                      ) / e1f(ji,jj)   +   e3f_0(ji,jj,jk)
                  END SELECT
               END DO
            END DO
         END DO
         !
         ii0 =  87   ;   ii1 =  87        ! Lombok Strait (e1v was modified)
         ij0 = 232   ;   ij1 = 233
         DO jk = 1, jpkm1
            DO jj = mj0(ij0), mj1(ij1)
               DO ji = mi0(ii0), mi1(ii1)
                  SELECT CASE ( pout )
                  CASE( 'V' )
                     pe3_out(ji,jj,jk) = 0.5_wp * vmask(ji,jj,jk)                                        &
                    &                    * (   e2t(ji,jj  ) * ( pe3_in(ji,jj  ,jk) - e3t_0(ji,jj  ,jk) ) &
                    &                    +     e2t(ji,jj+1) * ( pe3_in(ji,jj+1,jk) - e3t_0(ji,jj+1,jk) ) &
                    &                      ) / e2v(ji,jj)   +   e3v_0(ji,jj,jk)
                  END SELECT
               END DO
            END DO
         END DO
         !
         ii0 = 662   ;   ii1 = 662        ! Bab el Mandeb (e1v was modified)
         ij0 = 276   ;   ij1 = 276
         DO jk = 1, jpkm1
            DO jj = mj0(ij0), mj1(ij1)
               DO ji = mi0(ii0), mi1(ii1)
                  SELECT CASE ( pout )
                  CASE( 'V' )
                     pe3_out(ji,jj,jk) = 0.5_wp * vmask(ji,jj,jk)                                        &
                    &                    * (   e2t(ji,jj  ) * ( pe3_in(ji,jj  ,jk) - e3t_0(ji,jj  ,jk) ) &
                    &                    +     e2t(ji,jj+1) * ( pe3_in(ji,jj+1,jk) - e3t_0(ji,jj+1,jk) ) &
                    &                      ) / e2v(ji,jj)   +   e3v_0(ji,jj,jk)
                  END SELECT
               END DO
            END DO
         END DO
      ENDIF
   END SUBROUTINE dom_vvl_orca_fix

   SUBROUTINE dom_vvl_zdf( kt, p2dt ) 
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE dom_vvl_zdf  ***
      !!
      !! ** Purpose :   Do vertical thicknesses anomaly diffusion
      !!
      !! ** Method  :  
      !!
      !! ** Action  : 
      !!---------------------------------------------------------------------
      USE oce     , ONLY:   zwd => ua       , zws => va         ! (ua,va) used as 3D workspace
      !
      INTEGER                         , INTENT(in) ::   kt     ! ocean time-step index
      REAL(wp)                        , INTENT(in) ::   p2dt   ! time step
      !
      INTEGER  :: ji, jj, jk ! dummy loop indices
      REAL(wp) :: zr_tscale
      REAL(wp) :: za1, za2, za3, za4, zdiff
      REAL(wp), POINTER, DIMENSION(:,:,:) ::  zwi, zwt
      !!---------------------------------------------------------------------
      !
      IF( nn_timing == 1 )  CALL timing_start('dom_vvl_zdf')
      !
      zr_tscale = 0.5
      !
      CALL wrk_alloc( jpi, jpj, jpk, zwi, zwt) 
      !
      !

      zwt(:,:,1:jpk) = 1.e-10

      zwt(:,:,1) = 0.0_wp
!      DO jk = 2, jpkm1
!         DO jj = 1, jpj
!            DO ji = 1, jpi
!               zwt(ji,jj,jk) = zwt(ji,jj,jk-1) + (tilde_e3t_a(ji,jj,jk-1)+e3t_0(ji,jj,jk-1)) * tmask(ji,jj,jk-1)
!            END DO
!         END DO
!      END DO
      !
      !
      ! Set diffusivity (homogeneous to an inverse time scale)
      !
      DO jk = 2, jpkm1
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1
! Taper a little bit:
                za1 = tilde_e3t_n(ji,jj,jk-1)+e3t_0(ji,jj,jk-1)
                za2 = tilde_e3t_n(ji,jj,jk  )+e3t_0(ji,jj,jk  )
                za4 = 0.5_wp * (e3t_0(ji,jj,jk-1) + e3t_0(ji,jj,jk  ))
                za3 = 0.5_wp * (za1 + za2) 
                zdiff = ABS(za3-za4)/za4
                IF (zdiff>=0.8) THEN
                   zwt(ji,jj,jk) =  zr_tscale * MIN(zdiff,1._wp) * za3 / p2dt * tmask(ji,jj,jk)
!                   zwt(ji,jj,jk) =  dsm(ji,jj)/ht_0(ji,jj)*(1._wp-tanh((mbkt(ji,jj)+1-jk)*0.2))*tmask(ji,jj,jk)

                ELSE
                   zwt(ji,jj,jk) = 0.e0*tmask(ji,jj,jk)
                ENDIF
            END DO
         END DO
      END DO
      ! 
      !
      ! Diagonal, lower (i), upper (s)  (including the bottom boundary condition since avt is masked)
      DO jk = 1, jpkm1
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1   ! vector opt.
                zwi(ji,jj,jk) = - p2dt * zwt(ji,jj,jk  ) / fse3w(ji,jj,jk  )
                zws(ji,jj,jk) = - p2dt * zwt(ji,jj,jk+1) / fse3w(ji,jj,jk+1)
                zwd(ji,jj,jk) = 1._wp - zwi(ji,jj,jk) - zws(ji,jj,jk)
            END DO
         END DO
      END DO
      !
      !! Matrix inversion from the first level
      !!----------------------------------------------------------------------
      DO jj = 2, jpjm1
         DO ji = fs_2, fs_jpim1
            zwt(ji,jj,1) = zwd(ji,jj,1)
         END DO
      END DO
      DO jk = 2, jpkm1
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1
               zwt(ji,jj,jk) = zwd(ji,jj,jk) - zwi(ji,jj,jk) * zws(ji,jj,jk-1) / zwt(ji,jj,jk-1)
            END DO
         END DO
      END DO
      !         
      ! second recurrence:    Zk = Yk - Ik / Tk-1  Zk-1
      DO jk = 2, jpkm1
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1
               tilde_e3t_a(ji,jj,jk) = tilde_e3t_a(ji,jj,jk) - zwi(ji,jj,jk) / zwt(ji,jj,jk-1) * tilde_e3t_a(ji,jj,jk-1)
            END DO
         END DO
      END DO
      ! third recurrence:    Xk = (Zk - Sk Xk+1 ) / Tk
      DO jj = 2, jpjm1
         DO ji = fs_2, fs_jpim1
            tilde_e3t_a(ji,jj,jpkm1) = tilde_e3t_a(ji,jj,jpkm1) / zwt(ji,jj,jpkm1) * tmask(ji,jj,jpkm1)
         END DO
      END DO
      DO jk = jpk-2, 1, -1
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1
               tilde_e3t_a(ji,jj,jk) = ( tilde_e3t_a(ji,jj,jk) - zws(ji,jj,jk) * tilde_e3t_a(ji,jj,jk+1) )   &
                  &            / zwt(ji,jj,jk) * tmask(ji,jj,jk)
            END DO
         END DO
      END DO
      !
      CALL wrk_dealloc( jpi, jpj, jpk, zwi, zwt) 
      !
      IF( nn_timing == 1 )  CALL timing_stop('dom_vvl_zdf')
      !
   END SUBROUTINE dom_vvl_zdf

   SUBROUTINE dom_vvl_zdf2( kt, p2dt ) 
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE dom_vvl_zdf  ***
      !!
      !! ** Purpose :   Do vertical interface diffusion
      !!
      !! ** Method  :  
      !!
      !! ** Action  : 
      !!---------------------------------------------------------------------
      USE oce     , ONLY:   zwd => ua       , zws => va         ! (ua,va) used as 3D workspace
      !
      INTEGER                         , INTENT(in) ::   kt     ! ocean time-step index
      REAL(wp)                        , INTENT(in) ::   p2dt   ! time step
      !
      INTEGER  :: ji, jj, jk, kbot, kbotm1 ! dummy loop indices
      REAL(wp) :: zr_tscale
      REAL(wp) :: za1, za2, za3, za4, zdiff
      REAL(wp), POINTER, DIMENSION(:,:,:) ::  zwi, zwt, zw
      !!---------------------------------------------------------------------
      !
      IF( nn_timing == 1 )  CALL timing_start('dom_vvl_zdf')
      !
      zr_tscale = 0.5
      !
      CALL wrk_alloc( jpi, jpj, jpk, zwi, zwt, zw) 
      !
      ! Compute internal interfaces depths:
      zw(:,:,1) = 0._wp
      DO jk = 2, jpk
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1   ! vector opt.
               zw(ji,jj,jk) = zw(ji,jj,jk-1) + (tilde_e3t_a(ji,jj,jk-1)+e3t_0(ji,jj,jk-1))  * tmask(ji,jj,jk-1)
            END DO
         END DO
      END DO
      !
      ! Set diffusivities at interfaces
      zwt(:,:,:) = 0.00000001_wp * tmask(:,:,:)      
      zwt(:,:,1) = 0._wp
      ! 
      !
      ! Diagonal, lower (i), upper (s)  (including the bottom boundary condition since avt is masked)
      DO jk = 2, jpkm1
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1   ! vector opt.
                zwi(ji,jj,jk) = - 0.5_wp * p2dt * (zwt(ji,jj,jk-1) + zwt(ji,jj,jk  )) & 
                              &                   / fse3w(ji,jj,jk-1) / fse3t(ji,jj,jk-1) &
                              &                   * ht(ji,jj) 
                zws(ji,jj,jk) = - 0.5_wp * p2dt * (zwt(ji,jj,jk  ) + zwt(ji,jj,jk+1)) & 
                              &                   / fse3w(ji,jj,jk  ) / fse3t(ji,jj,jk  ) &
                              &                   * ht(ji,jj) 

                zwd(ji,jj,jk) = 1._wp - zwi(ji,jj,jk) - zws(ji,jj,jk)
            END DO
         END DO
      END DO
      ! Boundary conditions (Neumann)
      DO jj = 2, jpjm1
         DO ji = fs_2, fs_jpim1
            zwi(ji,jj,1) = 0._wp
            zws(ji,jj,1) = 0._wp
            zwd(ji,jj,1) = 1._wp
            zw (ji,jj,1) = 0._wp
            !
            zwd(ji,jj,2) = zwd(ji,jj,2) +  zwi(ji,jj,2)
            zwi(ji,jj,2) = 0._wp
!            zwi(ji,jj,2) = 0._wp
!            zws(ji,jj,2) = 0._wp
!            zwd(ji,jj,2) = 1._wp
         END DO
      END DO
      DO jj = 2, jpjm1
         DO ji = fs_2, fs_jpim1
            kbot   = mbkt(ji,jj) + 1
            kbotm1 = mbkt(ji,jj)
            zwi(ji,jj,kbot  ) = 0._wp
            zws(ji,jj,kbot  ) = 0._wp
            zwd(ji,jj,kbot  ) = 1._wp
            !
            zwd(ji,jj,kbotm1) = zwd(ji,jj,kbotm1) +  zws(ji,jj,kbotm1)
            zws(ji,jj,kbotm1) = 0._wp
!            zwi(ji,jj,kbotm1) = 0._wp
!            zws(ji,jj,kbotm1) = 0._wp
!            zwd(ji,jj,kbotm1) = 1._wp
         END DO
      END DO
      !
      DO jk = 2, jpkm1
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1
               zwd(ji,jj,jk) = zwd(ji,jj,jk) - zwi(ji,jj,jk) * zws(ji,jj,jk-1) / zwd(ji,jj,jk-1)
            END DO
         END DO
      END DO
      !
      DO jk = 2, jpk
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1    ! vector opt.
               zwi(ji,jj,jk) = zw(ji,jj,jk) - zwi(ji,jj,jk) / zwd(ji,jj,jk-1) *zwi(ji,jj,jk-1)
            END DO
         END DO
      END DO
      !
      DO jk = jpk-1, 2, -1
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1    ! vector opt.
               zw(ji,jj,jk) = ( zwi(ji,jj,jk) - zws(ji,jj,jk) * zw(ji,jj,jk+1) ) / zwd(ji,jj,jk)
            END DO
         END DO
      END DO
      !
      ! Revert to thicknesses anomalies:
      DO jk = 1, jpkm1
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1   ! vector opt.
               tilde_e3t_a(ji,jj,jk) = (zw(ji,jj,jk+1)-zw(ji,jj,jk)-e3t_0(ji,jj,jk))* tmask(ji,jj,jk)
            END DO
         END DO
      END DO
      !
      CALL wrk_dealloc( jpi, jpj, jpk, zwi, zwt, zw) 
      !
      IF( nn_timing == 1 )  CALL timing_stop('dom_vvl_zdf')
      !
   END SUBROUTINE dom_vvl_zdf2

   SUBROUTINE dom_vvl_regrid( kt )
      !!----------------------------------------------------------------------
      !!                ***  ROUTINE dom_vvl_regrid  ***
      !!                   
      !! ** Purpose :  Ensure "well-behaved" vertical grid
      !!
      !! ** Method  :  More or less adapted from references below.
      !!regrid
      !! ** Action  :  Ensure that thickness are above a given value, spaced enough
      !!               and revert to Eulerian coordinates near the bottom.      
      !!
      !! References :  Bleck, R. and S. Benjamin, 1993: Regional Weather Prediction
      !!               with a Model Combining Terrain-following and Isentropic
      !!               coordinates. Part I: Model Description. Monthly Weather Rev.,
      !!               121, 1770-1785.
      !!               Toy, M., 2011: Incorporating Condensational Heating into a
      !!               Nonhydrostatic Atmospheric Model Based on a Hybrid Isentropic-
      !!               Sigma Vertical Coordinate. Monthly Weather Rev., 139, 2940-2954.
      !!----------------------------------------------------------------------
      !! * Arguments
      INTEGER, INTENT( in )               :: kt         ! time step

      !! * Local declarations
      INTEGER                             :: ji, jj, jk ! dummy loop indices
      LOGICAL                             :: ll_chk_bot2top, ll_chk_top2bot, ll_lapdiff_cond
      LOGICAL                             :: ll_zdiff_cond, ll_blpdiff_cond
      INTEGER                             :: jkbot
      REAL(wp)                            :: zh_min, zh_0, zh2, zdiff, zh_max, ztmph, ztmpd
      REAL(wp)                            :: zufim1, zufi, zvfjm1, zvfj, dzmin_int, dzmin_surf
      REAL(wp)                            :: zh_new, zh_old, zh_bef, ztmp, ztmp1, z2dt, zh_up, zh_dwn
      REAL(wp)                            :: zeu2, zev2, zfrch_stp, zfrch_rel, zfrac_bot, zscal_bot
      REAL(wp)                            :: zhdiff, zhdiff2, zvdiff, zhlim, zhlim2, zvlim
      REAL(wp), POINTER, DIMENSION(:,:)   :: zdw
      REAL(wp), POINTER, DIMENSION(:,:,:) :: zwdw, zwdw_b
      !!----------------------------------------------------------------------

      IF( nn_timing == 1 )  CALL timing_start('dom_vvl_regrid')
      !
      CALL wrk_alloc( jpi, jpj, jpk, zwdw) 
      !
      ! Some user defined parameters below:
      ll_chk_bot2top   = .TRUE.
      ll_chk_top2bot   = .TRUE.
      dzmin_int  = 0.1_wp  ! Absolute minimum depth in the interior (in meters)
      dzmin_surf = 1.0_wp   ! Absolute minimum depth at the surface (in meters)
      zfrch_stp  = 0.5_wp   ! Maximum fractionnal thickness change in one time step (<= 1.)
      zfrch_rel  = 0.4_wp   ! Maximum relative thickness change in the vertical (<= 1.)
      zfrac_bot  = 0.05_wp  ! Fraction of bottom level allowed to change
      zscal_bot  = 2.0_wp   ! Depth lengthscale
      ll_zdiff_cond    = .TRUE.  ! Conditionnal vertical diffusion of interfaces
         zvdiff  = 0.1_wp   ! m
         zvlim   = 0.5_wp   ! max d2h/dh
      ll_lapdiff_cond  = .TRUE.  ! Conditionnal Laplacian diffusion of interfaces
         zhdiff  = 0.2_wp   ! ad.
         zhlim   = 0.03_wp  ! ad. max lap(z)*e1
      ll_blpdiff_cond  = .TRUE.  ! Conditionnal Bilaplacian diffusion of interfaces
         zhdiff2 = 0.2_wp   ! ad.
!         zhlim2  = 3.e-11_wp  !  max bilap(z)
         zhlim2  = 0.03_wp  ! ad. max bilap(z)*e1**3
      ! --------------------------------------------------------------------------------------- 
      !
      ! Set arrays determining maximum vertical displacement at the bottom:
      !--------------------------------------------------------------------
      IF ( kt==nit000 ) THEN
         DO jj = 2, jpjm1
            DO ji = 2, jpim1
               jk = MIN(mbkt(ji,jj), mbkt(ji+1,jj), mbkt(ji-1,jj), mbkt(ji,jj+1), mbkt(ji,jj-1))
               i_int_bot(ji,jj) = jk
            END DO
         END DO
         dsm(:,:) = REAL( i_int_bot(:,:), wp )   ;   CALL lbc_lnk(dsm(:,:),'T',1.)
         i_int_bot(:,:) = MAX( INT( dsm(:,:) ), 1 )

         CALL wrk_alloc( jpi, jpj, zdw )
         DO jj = 2, jpjm1
            DO ji = 2, jpim1
               zdw(ji,jj) = MAX(ABS(ht_0(ji,jj)-ht_0(ji+1,jj))*umask(ji  ,jj,1), &
                          &     ABS(ht_0(ji,jj)-ht_0(ji-1,jj))*umask(ji-1,jj,1), &
                          &     ABS(ht_0(ji,jj)-ht_0(ji,jj+1))*vmask(ji,jj  ,1), &
                          &     ABS(ht_0(ji,jj)-ht_0(ji,jj-1))*vmask(ji,jj-1,1)  )
                zdw(ji,jj) = MAX(zscal_bot * zdw(ji,jj), rsmall )
            END DO
         END DO
         CALL lbc_lnk( zdw(:,:), 'T', 1. )

         DO jj = 2, jpjm1
            DO ji = 2, jpim1
               dsm(ji,jj) = 1._wp/16._wp * (            zdw(ji-1,jj-1) + zdw(ji+1,jj-1)      &
                  &                           +         zdw(ji-1,jj+1) + zdw(ji+1,jj+1)      &
                  &                           + 2._wp*( zdw(ji  ,jj-1) + zdw(ji-1,jj  )      &
                  &                           +         zdw(ji+1,jj  ) + zdw(ji  ,jj+1) )    &
                  &                           + 4._wp*  zdw(ji  ,jj  )                       )
            END DO
         END DO         

         CALL lbc_lnk( dsm(:,:), 'T', 1. )
         CALL wrk_dealloc( jpi, jpj, zdw)         
     
         IF (ln_zps) THEN
            DO jj = 1, jpj
               DO ji = 1, jpi
                  jk = i_int_bot(ji,jj)
                  hsm(ji,jj) = zfrac_bot * e3w_1d(jk)
!                  dsm(ji,jj) = MAX(dsm(ji,jj), 0.1_wp*ht_0(ji,jj))
               END DO
            END DO
         ELSE
            DO jj = 1, jpj
               DO ji = 1, jpi
                  jk = i_int_bot(ji,jj)
                  hsm(ji,jj) = zfrac_bot * e3w_0(ji,jj,jk)
!                  dsm(ji,jj) = MAX(dsm(ji,jj), 0.1_wp*ht_0(ji,jj))
               END DO
            END DO
         ENDIF
      END IF

      ! Provisionnal interface depths:
      !-------------------------------
      zwdw(:,:,1) = 0.e0
      DO jj = 1, jpj
         DO ji = 1, jpi
            DO jk = 2, jpk
               zwdw(ji,jj,jk) =  zwdw(ji,jj,jk-1) + & 
                              & (tilde_e3t_a(ji,jj,jk-1)+e3t_0(ji,jj,jk-1)) * tmask(ji,jj,jk-1)
            END DO
         END DO
      END DO
      !
      ! Conditionnal horizontal Laplacian diffusion:
      !---------------------------------------------
      IF ( ll_lapdiff_cond ) THEN
         CALL wrk_alloc( jpi, jpj, jpk, zwdw_b)
         !
         zwdw_b(:,:,1) = 0._wp
         DO jj = 1, jpj
            DO ji = 1, jpi
               DO jk=2,jpk
                  zwdw_b(ji,jj,jk) =  zwdw_b(ji,jj,jk-1) + & 
                                   & (tilde_e3t_b(ji,jj,jk-1)+e3t_0(ji,jj,jk-1)) * tmask(ji,jj,jk-1)
               END DO
            END DO
         END DO
         !
         DO jk = 2, jpkm1
            DO jj = 1, jpjm1
               DO ji = 1, fs_jpim1   ! vector opt.                  
                  ua(ji,jj,jk) =  umask(ji,jj,jk) * re2u_e1u(ji,jj) &
                                  &  * ( zwdw_b(ji,jj,jk) - zwdw_b(ji+1,jj  ,jk) )
                  va(ji,jj,jk) =  vmask(ji,jj,jk) * re1v_e2v(ji,jj) & 
                                  &  * ( zwdw_b(ji,jj,jk) - zwdw_b(ji  ,jj+1,jk) )
               END DO
            END DO
         END DO
            
         DO jk = 2, jpkm1
            DO jj = 2, jpjm1
               DO ji = fs_2, fs_jpim1   ! vector opt.
                  ztmp1 = ( (ua(ji-1,jj  ,jk) - ua(ji,jj,jk))    &
                     &  +   (va(ji  ,jj-1,jk) - va(ji,jj,jk)) ) * r1_e12t(ji,jj)
                  zh2 = MAX(abs(ztmp1)-zhlim*SQRT(r1_e12t(ji,jj)), 0._wp)
                  ztmp = SIGN(zh2, ztmp1)
                  zeu2 = zhdiff * e12t(ji,jj)*e12t(ji,jj)/(e1t(ji,jj)*e1t(ji,jj) + e2t(ji,jj)*e2t(ji,jj))
                  zwdw(ji,jj,jk) = zwdw(ji,jj,jk) + zeu2 * ztmp *  tmask(ji,jj,jk)
               END DO
            END DO
         END DO         
         !
         CALL wrk_dealloc( jpi, jpj, jpk, zwdw_b)
         !
      ENDIF

      ! Conditionnal horizontal Bilaplacian diffusion:
      !-----------------------------------------------
      IF ( ll_blpdiff_cond ) THEN
         CALL wrk_alloc( jpi, jpj, jpk, zwdw_b)
         !
         zwdw_b(:,:,1) = 0._wp
         DO jj = 1, jpj
            DO ji = 1, jpi
               DO jk=2,jpk
                  zwdw_b(ji,jj,jk) =  zwdw_b(ji,jj,jk-1) + & 
                                   & (tilde_e3t_b(ji,jj,jk-1)+e3t_0(ji,jj,jk-1)) * tmask(ji,jj,jk-1)
               END DO
            END DO
         END DO
         !
         DO jk = 2, jpkm1
            DO jj = 1, jpjm1
               DO ji = 1, fs_jpim1   ! vector opt.                  
                  ua(ji,jj,jk) =  umask(ji,jj,jk) * re2u_e1u(ji,jj) &
                                  &  * ( zwdw_b(ji,jj,jk) - zwdw_b(ji+1,jj  ,jk) )
                  va(ji,jj,jk) =  vmask(ji,jj,jk) * re1v_e2v(ji,jj) & 
                                  &  * ( zwdw_b(ji,jj,jk) - zwdw_b(ji  ,jj+1,jk) )
               END DO
            END DO
         END DO
            
         DO jk = 2, jpkm1
            DO jj = 2, jpjm1
               DO ji = fs_2, fs_jpim1   ! vector opt.
                  zwdw_b(ji,jj,jk) = -( (ua(ji-1,jj  ,jk) - ua(ji,jj,jk))    &
                                &  +    (va(ji  ,jj-1,jk) - va(ji,jj,jk)) ) * r1_e12t(ji,jj)
               END DO
            END DO
         END DO   
         !
         CALL lbc_lnk( zwdw_b(:,:,:), 'T', 1. )      
         !
         DO jk = 2, jpkm1
            DO jj = 1, jpjm1
               DO ji = 1, fs_jpim1   ! vector opt.                  
                  ua(ji,jj,jk) =  umask(ji,jj,jk) * re2u_e1u(ji,jj) &
                                  &  * ( zwdw_b(ji,jj,jk) - zwdw_b(ji+1,jj  ,jk) )
                  va(ji,jj,jk) =  vmask(ji,jj,jk) * re1v_e2v(ji,jj) & 
                                  &  * ( zwdw_b(ji,jj,jk) - zwdw_b(ji  ,jj+1,jk) )
               END DO
            END DO
         END DO
         !   
         DO jk = 2, jpkm1
            DO jj = 2, jpjm1
               DO ji = fs_2, fs_jpim1   ! vector opt.
                  ztmp1 = ( (ua(ji-1,jj  ,jk) - ua(ji,jj,jk))    &
                     &  +   (va(ji  ,jj-1,jk) - va(ji,jj,jk)) ) * r1_e12t(ji,jj)
                  zh2 = MAX(abs(ztmp1)-zhlim2*SQRT(r1_e12t(ji,jj))*r1_e12t(ji,jj), 0._wp)
                  ztmp = SIGN(zh2, ztmp1)
                  zeu2 = zhdiff2 * e12t(ji,jj)*e12t(ji,jj) / 16._wp
                  zwdw(ji,jj,jk) = zwdw(ji,jj,jk) + zeu2 * ztmp *  tmask(ji,jj,jk)
               END DO
            END DO
         END DO   
         !
         CALL wrk_dealloc( jpi, jpj, jpk, zwdw_b)
         !
      ENDIF

      ! Conditionnal vertical diffusion:
      !---------------------------------
      IF ( ll_zdiff_cond ) THEN
         DO jk = 2, jpkm1
            DO jj = 2, jpjm1
               DO ji = fs_2, fs_jpim1   ! vector opt.    
                  ztmp  = -( (tilde_e3t_b(ji,jj,jk-1)+e3t_0(ji,jj,jk-1))*tmask(ji,jj,jk-1) & 
                            -(tilde_e3t_b(ji,jj,jk  )+e3t_0(ji,jj,jk  ))*tmask(ji,jj,jk  ) ) 
                  ztmp1 = 0.5_wp * ( tilde_e3t_b(ji,jj,jk-1) + e3t_0(ji,jj,jk-1) & 
                        &           +tilde_e3t_b(ji,jj,jk  ) + e3t_0(ji,jj,jk  ) )
                  zh2   = MAX(abs(ztmp)-zvlim*ztmp1, 0._wp)
                  ztmp  = SIGN(zh2, ztmp)
                  IF ((jk==mbkt(ji,jj)).AND.(ln_zps)) ztmp=0.e0
                  zwdw(ji,jj,jk) = zwdw(ji,jj,jk) + zvdiff * ztmp * tmask(ji,jj,jk)
               END DO
            END DO
         END DO
      ENDIF
      !
      ! Check grid from the bottom to the surface
      !------------------------------------------
      IF ( ll_chk_bot2top ) THEN
         DO jj = 2, jpjm1
            DO ji = 2, jpim1
               jkbot = mbkt(ji,jj)                   
               DO jk = jkbot,2,-1
                  !
                  zh_0   = e3t_0(ji,jj,jk)
                  zh_bef = MIN(tilde_e3t_b(ji,jj,jk) + zh_0, tilde_e3t_b(ji,jj,jk-1) + e3t_0(ji,jj,jk-1))
                  zh_old = zwdw(ji,jj,jk+1) - zwdw(ji,jj,jk)
                  zh_min = MIN(zh_0/3._wp, dzmin_int)
                  ! 
                  ! Set maximum and minimum vertical excursions   
                  ztmph = hsm(ji,jj)
                  ztmpd = dsm(ji,jj)
                  zh2   = ztmph * exp(-(gdepw_0(ji,jj,jk)-gdepw_0(ji,jj,i_int_bot(ji,jj)+1))/ztmpd)
                  zh2 = MAX(zh2,0.001_wp) ! Extend tolerance a bit for stability reasons (to be explored)
                  zdiff = cush_max(gdepw_0(ji,jj,jk)-zwdw(ji,jj,jk), zh2 )
                  zwdw(ji,jj,jk) = MAX(zwdw(ji,jj,jk), gdepw_0(ji,jj,jk) - zdiff)
                  zdiff = cush_max(zwdw(ji,jj,jk)-gdepw_0(ji,jj,jk), zh2 )
                  zwdw(ji,jj,jk) = MIN(zwdw(ji,jj,jk), gdepw_0(ji,jj,jk) + zdiff)
                  !
                  ! New layer thickness:
                  zh_new =  zwdw(ji,jj,jk+1) - zwdw(ji,jj,jk)
                  !
                  ! Ensure minimum layer thickness:                  
!                  zh_new = MAX((1._wp-zfrch_stp)*zh_bef, zh_new)
                  zh_new = cush(zh_new, zh_min)
                  !
                  ! Final flux:
                  zdiff = (zh_new - zh_old) * tmask(ji,jj,jk)
                  !
                  ! Limit thickness change in 1 time step: 
                  ztmp = MIN( ABS(zdiff), zfrch_stp*zh_bef )
                  zdiff = SIGN(ztmp, zh_new - zh_old)
                  zh_new = zdiff + zh_old
                  !
                  zwdw(ji,jj,jk) = zwdw(ji,jj,jk+1) - zh_new       
               END DO    
            END DO 
         END DO
      END IF    
      !
      ! Check grid from the surface to the bottom 
      !------------------------------------------ 
      IF ( ll_chk_top2bot ) THEN      
         DO jj = 2, jpjm1
            DO ji = 2, jpim1
               jkbot = mbkt(ji,jj)   
               DO jk = 1, jkbot-1
                  !
                  zh_0   = e3t_0(ji,jj,jk)
                  zh_bef = MIN(tilde_e3t_b(ji,jj,jk) + zh_0, tilde_e3t_b(ji,jj,jk+1) + e3t_0(ji,jj,jk+1))
                  zh_old = zwdw(ji,jj,jk+1) - zwdw(ji,jj,jk)
                  zh_min = MIN(zh_0/3._wp, dzmin_int)
                  !
                  zwdw(ji,jj,jk+1) = MAX(zwdw(ji,jj,jk+1), REAL(jk)*dzmin_surf)
                  !
                  ! New layer thickness:
                  zh_new =  zwdw(ji,jj,jk+1) - zwdw(ji,jj,jk)
                  !
                  ! Ensure minimum layer thickness:
!                  zh_new = MAX((1._wp-zfrch_stp)*zh_bef, zh_new)
                  zh_new = cush(zh_new, zh_min)
                  !
                  ! Final flux:
                  zdiff = (zh_new -zh_old) * tmask(ji,jj,jk)
                  ! 
                  ! Limit flux:                 
                  ztmp = MIN( ABS(zdiff), zfrch_stp*zh_bef )
                  zdiff = SIGN(ztmp, zh_new - zh_old)
                  zh_new = zdiff + zh_old
                  !
                  zwdw(ji,jj,jk+1) = zwdw(ji,jj,jk) + zh_new
               END DO
               !               
            END DO
         END DO
      ENDIF
      !
      DO jj = 2, jpjm1
         DO ji = 2, jpim1
            DO jk = 1, jpkm1
               tilde_e3t_a(ji,jj,jk) =  (zwdw(ji,jj,jk+1)-zwdw(ji,jj,jk)-e3t_0(ji,jj,jk)) * tmask(ji,jj,jk)
            END DO
         END DO
      END DO
      !
      CALL wrk_dealloc( jpi, jpj, jpk, zwdw ) 
      !
      IF( nn_timing == 1 )  CALL timing_stop('dom_vvl_regrid')
      !
   END SUBROUTINE dom_vvl_regrid

   FUNCTION cush(hin, hmin)  RESULT(hout)
      !!----------------------------------------------------------------------
      !!                 ***  FUNCTION cush  ***
      !!
      !! ** Purpose :   
      !!
      !! ** Method  :
      !!
      !!----------------------------------------------------------------------
      IMPLICIT NONE
      REAL(wp), INTENT(in) ::  hin, hmin
      REAL(wp)             ::  hout, zx, zh_cri
      !!----------------------------------------------------------------------
      zh_cri = 3._wp * hmin
      !
      IF ( hin<=0._wp ) THEN
         hout = hmin
      !
      ELSEIF ( (hin>0._wp).AND.(hin<=zh_cri) ) THEN
         zx = hin/zh_cri
         hout = hmin * (1._wp + zx + zx*zx)      
      !
      ELSEIF ( hin>zh_cri ) THEN
         hout = hin
      !
      ENDIF
      !
   END FUNCTION cush

   FUNCTION cush_max(hin, hmax)  RESULT(hout)
      !!----------------------------------------------------------------------
      !!                 ***  FUNCTION cush  ***
      !!
      !! ** Purpose :   
      !!
      !! ** Method  :
      !!
      !!----------------------------------------------------------------------
      IMPLICIT NONE
      REAL(wp), INTENT(in) ::  hin, hmax
      REAL(wp)             ::  hout, hmin, zx, zh_cri
      !!----------------------------------------------------------------------
      hmin = 0.1_wp * hmax
      zh_cri = 3._wp * hmin
      !
      IF ( (hin>=(hmax-zh_cri)).AND.(hin<=(hmax-hmin))) THEN
         zx = (hmax-hin)/zh_cri
         hout = hmax - hmin * (1._wp + zx + zx*zx)
         !
      ELSEIF ( hin>(hmax-zh_cri) ) THEN
         hout = hmax - hmin
         !
      ELSE
         hout = hin
         !
      ENDIF
      !
   END FUNCTION cush_max

   SUBROUTINE dom_vvl_adv_fct( kt, pta, uin, vin )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE dom_vvl_adv_fct  ***
      !! 
      !! **  Purpose :  Do thickness advection
      !!
      !! **  Method  :  FCT scheme to ensure positivity 
      !!
      !! **  Action  : - Update pta thickness tendency and diffusive fluxes
      !!               - this is the total trend, hence it does include sea level motions
      !!----------------------------------------------------------------------
      !
      INTEGER, INTENT( in ) ::   kt   ! ocean time-step index
      REAL(wp), DIMENSION(jpi,jpj,jpk), INTENT(inout) ::   pta    ! thickness baroclinic trend 
      REAL(wp), DIMENSION(jpi,jpj,jpk), INTENT(inout) ::   uin, vin  ! input velocities
      !
      INTEGER  ::   ji, jj, jk, ib, ib_bdy               ! dummy loop indices  
      INTEGER  ::   ikbu, ikbv, ibot
      REAL(wp) ::   z2dtt, zbtr, ztra        ! local scalar
      REAL(wp) ::   zdi, zdj, zmin           !   -      -
      REAL(wp) ::   zfp_ui, zfp_vj           !   -      -
      REAL(wp) ::   zfm_ui, zfm_vj           !   -      -
      REAL(wp) ::   zfp_hi, zfp_hj           !   -      -
      REAL(wp) ::   zfm_hi, zfm_hj           !   -      -
      REAL(wp) ::   ztout,  ztin, zfac       !   -      -
      REAL(wp), POINTER, DIMENSION(:,:,:) :: zwx, zwy, zwi
      !!----------------------------------------------------------------------
      !
      IF( nn_timing == 1 )  CALL timing_start('dom_vvl_adv_fct')
      !
      CALL wrk_alloc( jpi, jpj, jpk, zwx, zwy, zwi)
      !
      ! 1. Initializations
      ! ------------------
      !
      IF( neuler == 0 .AND. kt == nit000 ) THEN
         z2dtt =  rdt
      ELSE
         z2dtt = 2.0_wp * rdt
      ENDIF
      !
      zwi(:,:,:) = 0.e0
      zwx(:,:,:) = 0.e0 
      zwy(:,:,:) = 0.e0
      !
      !
      ! 2. upstream advection with initial mass fluxes & intermediate update
      ! --------------------------------------------------------------------
      IF ( ll_shorizd ) THEN
         DO jk = 1, jpkm1
            DO jj = 1, jpjm1
               DO ji = 1, fs_jpim1   ! vector opt.
                  !
                  zfp_hi = MAX(hu_b(ji,jj) - fsdepw_b(ji  ,jj  ,jk), 0._wp)
                  zfp_hi = MIN(zfp_hi, fse3t_b(ji  ,jj  ,jk))
                  zfp_hi = 0.5_wp *(zfp_hi + SIGN(zfp_hi, zfp_hi-hsmall) ) 
                  !
                  zfm_hi = MAX(hu_b(ji,jj) - fsdepw_b(ji+1,jj  ,jk), 0._wp)
                  zfm_hi = MIN(zfm_hi, fse3t_b(ji+1,jj  ,jk))
                  zfm_hi = 0.5_wp *(zfm_hi + SIGN(zfm_hi, zfm_hi-hsmall) ) 
                  ! 
                  zfp_hj = MAX(hv_b(ji,jj) - fsdepw_b(ji  ,jj  ,jk), 0._wp)
                  zfp_hj = MIN(zfp_hj, fse3t_b(ji  ,jj  ,jk))
                  zfp_hj = 0.5_wp *(zfp_hj + SIGN(zfp_hj, zfp_hj-hsmall) ) 
                  !
                  zfm_hj = MAX(hv_b(ji,jj) - fsdepw_b(ji  ,jj+1,jk), 0._wp)
                  zfm_hj = MIN(zfm_hj, fse3t_b(ji  ,jj+1,jk))
                  zfm_hj = 0.5_wp *(zfm_hj + SIGN(zfm_hj, zfm_hj-hsmall) )
                  !
                  zfp_ui = uin(ji,jj,jk) + ABS( uin(ji,jj,jk) )
                  zfm_ui = uin(ji,jj,jk) - ABS( uin(ji,jj,jk) )
                  zfp_vj = vin(ji,jj,jk) + ABS( vin(ji,jj,jk) )
                  zfm_vj = vin(ji,jj,jk) - ABS( vin(ji,jj,jk) )
                  zwx(ji,jj,jk) = 0.5 * e2u(ji,jj) * ( zfp_ui * zfp_hi + zfm_ui * zfm_hi ) * umask(ji,jj,jk)
                  zwy(ji,jj,jk) = 0.5 * e1v(ji,jj) * ( zfp_vj * zfp_hj + zfm_vj * zfm_hj ) * vmask(ji,jj,jk)
               END DO
            END DO
         END DO
      ELSE
         DO jk = 1, jpkm1
            DO jj = 1, jpjm1
               DO ji = 1, fs_jpim1   ! vector opt.
                  !
                  zfp_hi = fse3t_b(ji  ,jj  ,jk)
                  zfm_hi = fse3t_b(ji+1,jj  ,jk)             
                  zfp_hj = fse3t_b(ji  ,jj  ,jk)
                  zfm_hj = fse3t_b(ji  ,jj+1,jk)
                  !
                  zfp_ui = uin(ji,jj,jk) + ABS( uin(ji,jj,jk) )
                  zfm_ui = uin(ji,jj,jk) - ABS( uin(ji,jj,jk) )
                  zfp_vj = vin(ji,jj,jk) + ABS( vin(ji,jj,jk) )
                  zfm_vj = vin(ji,jj,jk) - ABS( vin(ji,jj,jk) )
                  zwx(ji,jj,jk) = 0.5 * e2u(ji,jj) * ( zfp_ui * zfp_hi + zfm_ui * zfm_hi ) * umask(ji,jj,jk)
                  zwy(ji,jj,jk) = 0.5 * e1v(ji,jj) * ( zfp_vj * zfp_hj + zfm_vj * zfm_hj ) * vmask(ji,jj,jk)
               END DO
            END DO
         END DO
      ENDIF

      ! total advective trend
      DO jk = 1, jpkm1
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1   ! vector opt.
               zbtr = r1_e12t(ji,jj)
               ! total intermediate advective trends
               ztra = - zbtr * (  zwx(ji,jj,jk) - zwx(ji-1,jj  ,jk  )  &
                  &             + zwy(ji,jj,jk) - zwy(ji  ,jj-1,jk  )  )
               !
               ! update and guess with monotonic sheme
               pta(ji,jj,jk) =   pta(ji,jj,jk) + ztra
               zwi(ji,jj,jk) = (fse3t_b(ji,jj,jk) + z2dtt * ztra ) * tmask(ji,jj,jk)
            END DO
         END DO
      END DO

      CALL lbc_lnk( zwi, 'T', 1. )  

#if defined key_bdy
         DO ib_bdy=1, nb_bdy
            DO ib = 1, idx_bdy(ib_bdy)%nblenrim(1)
               ji = idx_bdy(ib_bdy)%nbi(ib,1)
               jj = idx_bdy(ib_bdy)%nbj(ib,1)
               DO jk = 1, jpkm1  
                  zwi(ji,jj,jk) = fse3t_a(ji,jj,jk)
               END DO
            END DO 
         END DO
#endif

      IF ( ln_vvl_dbg ) THEN
         zmin = MINVAL( zwi(:,:,:), mask = tmask(:,:,:) == 1.e0 ) 
         IF( lk_mpp )   CALL mpp_min( zmin )
         IF( zmin < 0._wp) THEN
            IF(lwp) CALL ctl_warn('vvl_adv: CFL issue here')
            IF(lwp) WRITE(numout,*) zmin
         ENDIF
      ENDIF

      ! 3. antidiffusive flux : high order minus low order
      ! --------------------------------------------------
      ! antidiffusive flux on i and j
      DO jk = 1, jpkm1
         DO jj = 1, jpjm1
            DO ji = 1, fs_jpim1   ! vector opt.
               zwx(ji,jj,jk) = (e2u(ji,jj) * uin(ji,jj,jk) * fse3u_n(ji,jj,jk) &
                                & - zwx(ji,jj,jk)) * umask(ji,jj,jk)
               zwy(ji,jj,jk) = (e1v(ji,jj) * vin(ji,jj,jk) * fse3v_n(ji,jj,jk) & 
                                & - zwy(ji,jj,jk)) * vmask(ji,jj,jk)
               !
               ! Update advective fluxes
               un_td(ji,jj,jk) = un_td(ji,jj,jk) - zwx(ji,jj,jk)
               vn_td(ji,jj,jk) = vn_td(ji,jj,jk) - zwy(ji,jj,jk)
            END DO
         END DO
      END DO
      
      CALL lbc_lnk( zwx, 'U', -1. )   ;   CALL lbc_lnk( zwy, 'V', -1. )         ! Lateral boundary conditions

      ! 4. monotonicity algorithm
      ! -------------------------
      CALL nonosc_2d( fse3t_b(:,:,:), zwx, zwy, zwi, z2dtt )

      ! 5. final trend with corrected fluxes
      ! ------------------------------------
      !
      ! Update advective fluxes
      un_td(:,:,:) = (un_td(:,:,:) + zwx(:,:,:))*umask(:,:,:)
      vn_td(:,:,:) = (vn_td(:,:,:) + zwy(:,:,:))*vmask(:,:,:)
      !
      DO jk = 1, jpkm1
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1   ! vector opt.  
               !
               zbtr = r1_e12t(ji,jj)
               ztra = - zbtr * (  zwx(ji,jj,jk) - zwx(ji-1,jj  ,jk  )   &
                  &             + zwy(ji,jj,jk) - zwy(ji  ,jj-1,jk  ) )
               ! add them to the general tracer trends
               pta(ji,jj,jk) = pta(ji,jj,jk) + ztra
            END DO
         END DO
      END DO
      !
      CALL wrk_dealloc( jpi, jpj, jpk, zwx, zwy, zwi)
      !
      IF( nn_timing == 1 )  CALL timing_stop('dom_vvl_adv_fct')
      !
   END SUBROUTINE dom_vvl_adv_fct

   SUBROUTINE dom_vvl_ups_cor( kt, pta, uin, vin ) 
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE dom_vvl_adv_fct  ***
      !! 
      !! **  Purpose :  Correct for addionnal barotropic fluxes 
      !!                in the upstream direction
      !!
      !! **  Method  :   
      !!
      !! **  Action  : - Update diffusive fluxes uin, vin
      !!               - Remove divergence from thickness tendency
      !!----------------------------------------------------------------------
      INTEGER, INTENT( in ) ::   kt   ! ocean time-step index
      REAL(wp), DIMENSION(jpi,jpj,jpk), INTENT(inout) ::   pta       ! thickness baroclinic trend 
      REAL(wp), DIMENSION(jpi,jpj,jpk), INTENT(inout) ::   uin, vin  ! input fluxes
      INTEGER  ::   ji, jj, jk               ! dummy loop indices  
      INTEGER  ::   ikbu, ikbv, ibot
      REAL(wp) ::   zbtr, ztra               ! local scalar
      REAL(wp) ::   zdi, zdj                 !   -      -
      REAL(wp) ::   zfp_hi, zfp_hj           !   -      -
      REAL(wp) ::   zfm_hi, zfm_hj           !   -      -
      REAL(wp) ::   zfp_ui, zfp_vj           !   -      -
      REAL(wp) ::   zfm_ui, zfm_vj           !   -      -
      REAL(wp), POINTER, DIMENSION(:,:)   :: zbu, zbv, zhu_b, zhv_b
      REAL(wp), POINTER, DIMENSION(:,:,:) :: zwx, zwy
      !!----------------------------------------------------------------------
      !
      IF( nn_timing == 1 )  CALL timing_start('dom_vvl_ups_cor')
      !
      CALL wrk_alloc( jpi, jpj, zhu_b, zhv_b, zbu, zbv)
      CALL wrk_alloc( jpi, jpj, jpk, zwx, zwy)
      !
      ! Compute barotropic flux difference:
      zbu(:,:) = 0.e0
      zbv(:,:) = 0.e0
      DO jj = 1, jpj
         DO ji = 1, jpi   ! vector opt.
            DO jk = 1, jpkm1
               zbu(ji,jj) = zbu(ji,jj) - uin(ji,jj,jk) * umask(ji,jj,jk)
               zbv(ji,jj) = zbv(ji,jj) - vin(ji,jj,jk) * vmask(ji,jj,jk)
            END DO
         END DO
      ENDDO

      ! Compute upstream depths:
      zhu_b(:,:) = 0.e0
      zhv_b(:,:) = 0.e0

      IF ( ll_shorizd ) THEN
         ! Correct bottom value
         ! considering "shelf horizon depth"
         DO jj = 1, jpjm1
            DO ji = 1, jpim1   ! vector opt.
               zdi = 0.5_wp + 0.5_wp * SIGN(1._wp, zbu(ji,jj))
               zdj = 0.5_wp + 0.5_wp * SIGN(1._wp, zbv(ji,jj))
               DO jk=1, jpkm1
                  zfp_hi = MAX(hu_b(ji,jj) - fsdepw_b(ji  ,jj  ,jk), 0._wp)
                  zfp_hi = MIN(zfp_hi, fse3t_b(ji  ,jj  ,jk))
                  zfp_hi = 0.5_wp *(zfp_hi + SIGN(zfp_hi, zfp_hi-hsmall) ) 
                  !
                  zfm_hi = MAX(hu_b(ji,jj) - fsdepw_b(ji+1,jj  ,jk), 0._wp)
                  zfm_hi = MIN(zfm_hi, fse3t_b(ji+1,jj  ,jk))
                  zfm_hi = 0.5_wp *(zfm_hi + SIGN(zfm_hi, zfm_hi-hsmall) )
                  ! 
                  zfp_hj = MAX(hv_b(ji,jj) - fsdepw_b(ji  ,jj  ,jk), 0._wp)
                  zfp_hj = MIN(zfp_hj, fse3t_b(ji  ,jj  ,jk))
                  zfp_hj = 0.5_wp *(zfp_hj + SIGN(zfp_hj, zfp_hj-hsmall) ) 
                  !
                  zfm_hj = MAX(hv_b(ji,jj) - fsdepw_b(ji  ,jj+1,jk), 0._wp)
                  zfm_hj = MIN(zfm_hj, fse3t_b(ji  ,jj+1,jk))
                  zfm_hj = 0.5_wp *(zfm_hj + SIGN(zfm_hj, zfm_hj-hsmall) )
                  !
                  zhu_b(ji,jj) = zhu_b(ji,jj) + (         zdi  * zfp_hi &
                               &                 + (1._wp-zdi) * zfm_hi &
                               &                ) * umask(ji,jj,jk)
                  zhv_b(ji,jj) = zhv_b(ji,jj) + (         zdj  * zfp_hj &
                               &                 + (1._wp-zdj) * zfm_hj &
                               &                ) * vmask(ji,jj,jk)
               END DO
            END DO
         END DO
      ELSE
         DO jj = 1, jpjm1
            DO ji = 1, jpim1   ! vector opt.
               zdi = 0.5_wp + 0.5_wp * SIGN(1._wp, zbu(ji,jj)) 
               zdj = 0.5_wp + 0.5_wp * SIGN(1._wp, zbv(ji,jj))
               DO jk = 1, jpkm1
                  zfp_hi = fse3t_b(ji  ,jj  ,jk)
                  zfm_hi = fse3t_b(ji+1,jj  ,jk)
                  zfp_hj = fse3t_b(ji  ,jj  ,jk)
                  zfm_hj = fse3t_b(ji  ,jj+1,jk)
                  !
                  zhu_b(ji,jj) = zhu_b(ji,jj) + (          zdi  * zfp_hi &
                               &                  + (1._wp-zdi) * zfm_hi &
                               &                ) * umask(ji,jj,jk)
                  !
                  zhv_b(ji,jj) = zhv_b(ji,jj) + (          zdj  * zfp_hj &
                               &                  + (1._wp-zdj) * zfm_hj &
                               &                ) * vmask(ji,jj,jk)
               END DO
            END DO
         END DO
      ENDIF

      ! Corrective barotropic velocity (times hor. scale factor)
      zbu(:,:) = zbu(:,:)/ (zhu_b(:,:)*umask(:,:,1)+1._wp-umask(:,:,1))
      zbv(:,:) = zbv(:,:)/ (zhv_b(:,:)*vmask(:,:,1)+1._wp-vmask(:,:,1))

      CALL lbc_lnk( zbu(:,:), 'U', -1. )
      CALL lbc_lnk( zbv(:,:), 'V', -1. )
      
      ! Set corrective fluxes in upstream direction:
      !
      zwx(:,:,:) = 0.e0
      zwy(:,:,:) = 0.e0
      IF ( ll_shorizd ) THEN
         DO jj = 1, jpjm1
            DO ji = 1, fs_jpim1   ! vector opt.
               ! upstream scheme
               zfp_ui = zbu(ji,jj) + ABS( zbu(ji,jj) )
               zfm_ui = zbu(ji,jj) - ABS( zbu(ji,jj) )
               zfp_vj = zbv(ji,jj) + ABS( zbv(ji,jj) )
               zfm_vj = zbv(ji,jj) - ABS( zbv(ji,jj) )
               DO jk = 1, jpkm1
                  zfp_hi = MAX(hu_b(ji,jj) - fsdepw_b(ji  ,jj  ,jk), 0._wp)
                  zfp_hi = MIN(fse3t_b(ji  ,jj  ,jk), zfp_hi)
                  zfp_hi = 0.5_wp *(zfp_hi + SIGN(zfp_hi, zfp_hi-hsmall) ) 
                  !
                  zfm_hi = MAX(hu_b(ji,jj) - fsdepw_b(ji+1,jj  ,jk), 0._wp)
                  zfm_hi = MIN(fse3t_b(ji+1,jj  ,jk), zfm_hi)
                  zfm_hi = 0.5_wp *(zfm_hi + SIGN(zfm_hi, zfm_hi-hsmall) ) 
                  !
                  zfp_hj = MAX(hv_b(ji,jj) - fsdepw_b(ji  ,jj  ,jk), 0._wp)
                  zfp_hj = MIN(fse3t_b(ji  ,jj  ,jk), zfp_hj) 
                  zfp_hj = 0.5_wp *(zfp_hj + SIGN(zfp_hj, zfp_hj-hsmall) ) 
                  !
                  zfm_hj = MAX(hv_b(ji,jj) - fsdepw_b(ji  ,jj+1,jk), 0._wp)
                  zfm_hj = MIN(fse3t_b(ji  ,jj+1,jk), zfm_hj)
                  zfm_hj = 0.5_wp *(zfm_hj + SIGN(zfm_hj, zfm_hj-hsmall) ) 
                  !
                  zwx(ji,jj,jk) = 0.5 * ( zfp_ui * zfp_hi + zfm_ui * zfm_hi ) * umask(ji,jj,jk)
                  zwy(ji,jj,jk) = 0.5 * ( zfp_vj * zfp_hj + zfm_vj * zfm_hj ) * vmask(ji,jj,jk)
               END DO
            END DO
         END DO
      ELSE
         DO jj = 1, jpjm1
            DO ji = 1, fs_jpim1   ! vector opt.
               ! upstream scheme
               zfp_ui = zbu(ji,jj) + ABS( zbu(ji,jj) )
               zfm_ui = zbu(ji,jj) - ABS( zbu(ji,jj) )
               zfp_vj = zbv(ji,jj) + ABS( zbv(ji,jj) )
               zfm_vj = zbv(ji,jj) - ABS( zbv(ji,jj) )
               DO jk = 1, jpkm1
                  zfp_hi = fse3t_b(ji  ,jj  ,jk)
                  zfm_hi = fse3t_b(ji+1,jj  ,jk)
                  zfp_hj = fse3t_b(ji  ,jj  ,jk)
                  zfm_hj = fse3t_b(ji  ,jj+1,jk)
                  !
                  zwx(ji,jj,jk) = 0.5 * ( zfp_ui * zfp_hi + zfm_ui * zfm_hi ) * umask(ji,jj,jk)
                  zwy(ji,jj,jk) = 0.5 * ( zfp_vj * zfp_hj + zfm_vj * zfm_hj ) * vmask(ji,jj,jk)
               END DO
            END DO
         END DO
      ENDIF
      CALL lbc_lnk( zwx, 'U', -1. )   ;   CALL lbc_lnk( zwy, 'V', -1. )         ! Lateral boundary conditions

      uin(:,:,:) = uin(:,:,:) + zwx(:,:,:)
      vin(:,:,:) = vin(:,:,:) + zwy(:,:,:)
      !
      ! Update trend with corrective fluxes:
      DO jk = 1, jpkm1
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1   ! vector opt.  
               !
               zbtr = r1_e12t(ji,jj)
               ! total advective trends
               ztra = - zbtr * (  zwx(ji,jj,jk) - zwx(ji-1,jj  ,jk  )   &
                  &             + zwy(ji,jj,jk) - zwy(ji  ,jj-1,jk  ) )
               ! add them to the general tracer trends
               pta(ji,jj,jk) = pta(ji,jj,jk) + ztra
            END DO
         END DO
      END DO
      !
      CALL wrk_dealloc( jpi, jpj, zhu_b, zhv_b, zbu, zbv)
      CALL wrk_dealloc( jpi, jpj, jpk, zwx, zwy)
      !
      IF( nn_timing == 1 )  CALL timing_stop('dom_vvl_ups_cor')
      !
   END SUBROUTINE dom_vvl_ups_cor

   SUBROUTINE nonosc_2d( pbef, paa, pbb, paft, p2dt )
      !!---------------------------------------------------------------------
      !!                    ***  ROUTINE nonosc_2d  ***
      !!     
      !! **  Purpose :   compute monotonic thickness fluxes from the upstream 
      !!       scheme and the before field by a nonoscillatory algorithm 
      !!
      !! **  Method  :   ... ???
      !!       warning : pbef and paft must be masked, but the boundaries
      !!       conditions on the fluxes are not necessary zalezak (1979)
      !!       drange (1995) multi-dimensional forward-in-time and upstream-
      !!       in-space based differencing for fluid
      !!----------------------------------------------------------------------
      !
      !!----------------------------------------------------------------------
      REAL(wp)                         , INTENT(in   ) ::   p2dt            ! vertical profile of tracer time-step
      REAL(wp), DIMENSION (jpi,jpj,jpk), INTENT(in   ) ::   pbef, paft      ! before & after field
      REAL(wp), DIMENSION (jpi,jpj,jpk), INTENT(inout) ::   paa, pbb        ! monotonic fluxes in the 3 directions
      !
      INTEGER ::   ji, jj, jk   ! dummy loop indices
      REAL(wp) ::   zpos, zneg, zbt, za, zb, zc, zbig, zrtrn, z2dtt   ! local scalars
      REAL(wp) ::   zau, zbu, zcu, zav, zbv, zcv, zup, zdo            !   -      -
      REAL(wp) ::   zupip1, zupim1, zupjp1, zupjm1, zupb, zupa
      REAL(wp) ::   zdoip1, zdoim1, zdojp1, zdojm1, zdob, zdoa
      REAL(wp), POINTER, DIMENSION(:,:,:) :: zbetup, zbetdo, zbup, zbdo
      !!----------------------------------------------------------------------
      !
      IF( nn_timing == 1 )  CALL timing_start('nonosc2')
      !
      CALL wrk_alloc( jpi, jpj, jpk, zbetup, zbetdo, zbup, zbdo )
      !

      zbig  = 1.e+40_wp
      zrtrn = 1.e-15_wp
      zbetup(:,:,jpk) = 0._wp   ;   zbetdo(:,:,jpk) = 0._wp


      ! Search local extrema
      ! --------------------
      ! max/min of pbef & paft with large negative/positive value (-/+zbig) inside land
      zbup = MAX( pbef * tmask - zbig * ( 1.e0 - tmask ),   &
         &        paft * tmask - zbig * ( 1.e0 - tmask )  )
      zbdo = MIN( pbef * tmask + zbig * ( 1.e0 - tmask ),   &
         &        paft * tmask + zbig * ( 1.e0 - tmask )  )

      DO jk = 1, jpkm1
         z2dtt = p2dt
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1   ! vector opt.

               ! search maximum in neighbourhood
               zup = MAX(  zbup(ji  ,jj  ,jk  ),   &
                  &        zbup(ji-1,jj  ,jk  ), zbup(ji+1,jj  ,jk  ),   &
                  &        zbup(ji  ,jj-1,jk  ), zbup(ji  ,jj+1,jk  ))

               ! search minimum in neighbourhood
               zdo = MIN(  zbdo(ji  ,jj  ,jk  ),   &
                  &        zbdo(ji-1,jj  ,jk  ), zbdo(ji+1,jj  ,jk  ),   &
                  &        zbdo(ji  ,jj-1,jk  ), zbdo(ji  ,jj+1,jk  ))

               ! positive part of the flux
               zpos = MAX( 0., paa(ji-1,jj  ,jk  ) ) - MIN( 0., paa(ji  ,jj  ,jk  ) )   &
                  & + MAX( 0., pbb(ji  ,jj-1,jk  ) ) - MIN( 0., pbb(ji  ,jj  ,jk  ) )   

               ! negative part of the flux
               zneg = MAX( 0., paa(ji  ,jj  ,jk  ) ) - MIN( 0., paa(ji-1,jj  ,jk  ) )   &
                  & + MAX( 0., pbb(ji  ,jj  ,jk  ) ) - MIN( 0., pbb(ji  ,jj-1,jk  ) )

               ! up & down beta terms
               zbt = e1t(ji,jj) * e2t(ji,jj) / z2dtt
               zbetup(ji,jj,jk) = ( zup            - paft(ji,jj,jk) ) / ( zpos + zrtrn ) * zbt
               zbetdo(ji,jj,jk) = ( paft(ji,jj,jk) - zdo            ) / ( zneg + zrtrn ) * zbt
            END DO
         END DO
      END DO

      CALL lbc_lnk( zbetup, 'T', 1. )   ;   CALL lbc_lnk( zbetdo, 'T', 1. )   ! lateral boundary cond. (unchanged sign)

      ! 3. monotonic flux in the i & j direction (paa & pbb)
      ! ----------------------------------------
      DO jk = 1, jpkm1
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1   ! vector opt.
               zau = MIN( 1.e0, zbetdo(ji,jj,jk), zbetup(ji+1,jj,jk) )
               zbu = MIN( 1.e0, zbetup(ji,jj,jk), zbetdo(ji+1,jj,jk) )
               zcu =       ( 0.5  + SIGN( 0.5 , paa(ji,jj,jk) ) )
               paa(ji,jj,jk) = paa(ji,jj,jk) * ( zcu * zau + ( 1.e0 - zcu) * zbu )

               zav = MIN( 1.e0, zbetdo(ji,jj,jk), zbetup(ji,jj+1,jk) )
               zbv = MIN( 1.e0, zbetup(ji,jj,jk), zbetdo(ji,jj+1,jk) )
               zcv =       ( 0.5  + SIGN( 0.5 , pbb(ji,jj,jk) ) )
               pbb(ji,jj,jk) = pbb(ji,jj,jk) * ( zcv * zav + ( 1.e0 - zcv) * zbv )
            END DO
         END DO
      END DO
      CALL lbc_lnk( paa, 'U', -1. )   ;   CALL lbc_lnk( pbb, 'V', -1. )   ! lateral boundary condition (changed sign)
      !
      CALL wrk_dealloc( jpi, jpj, jpk, zbetup, zbetdo, zbup, zbdo )
      !
      IF( nn_timing == 1 )  CALL timing_stop('nonosc2')
      !
   END SUBROUTINE nonosc_2d

   !!======================================================================
END MODULE domvvl
